# -*- coding: utf-8 -*-
"""
Tagulous test: Utils

Modules tested:
    tagulous.utils
"""
import unittest

from django.test import TestCase

from tagulous import utils as tag_utils


try:
    from unidecode import unidecode
except ImportError:
    unidecode = None


# ##############################################################################
# ###### utils.strip_split()
# ##############################################################################


class UtilsSplitStripTest(TestCase):
    "Test utils.split_strip"

    def test_empty(self):
        split = tag_utils.split_strip(None)
        self.assertEqual(split, [])
        split = tag_utils.split_strip("")
        self.assertEqual(split, [])

    def test_spaceless(self):
        split = tag_utils.split_strip("adam,brian")
        self.assertEqual(len(split), 2)
        self.assertEqual(split[0], "adam")
        self.assertEqual(split[1], "brian")

    def test_spaced(self):
        split = tag_utils.split_strip("  adam  ,  brian  ")
        self.assertEqual(len(split), 2)
        self.assertEqual(split[0], "adam")
        self.assertEqual(split[1], "brian")


# ##############################################################################
# ###### utils.parse_tags
# ##############################################################################


class UtilsParseTagsTest(TestCase):
    "Test utils.parse_tags"

    def test_commas(self):
        tags = tag_utils.parse_tags("adam,brian,chris")
        self.assertEqual(len(tags), 3)
        self.assertEqual(tags[0], "adam")
        self.assertEqual(tags[1], "brian")
        self.assertEqual(tags[2], "chris")

    def test_spaces(self):
        tags = tag_utils.parse_tags("adam brian chris")
        self.assertEqual(len(tags), 3)
        self.assertEqual(tags[0], "adam")
        self.assertEqual(tags[1], "brian")
        self.assertEqual(tags[2], "chris")

    def test_trailing_comma(self):
        tags = tag_utils.parse_tags("adam,brian,")
        self.assertEqual(len(tags), 2)
        self.assertEqual(tags[0], "adam")
        self.assertEqual(tags[1], "brian")

    def test_trailing_space(self):
        tags = tag_utils.parse_tags("adam brian ")
        self.assertEqual(len(tags), 2)
        self.assertEqual(tags[0], "adam")
        self.assertEqual(tags[1], "brian")

    def test_commas_and_spaces(self):
        tags = tag_utils.parse_tags("adam, brian chris")
        self.assertEqual(len(tags), 2)
        self.assertEqual(tags[0], "adam")
        self.assertEqual(tags[1], "brian chris")

    def test_commas_over_spaces(self):
        tags = tag_utils.parse_tags("adam brian  ,  chris")
        self.assertEqual(len(tags), 2)
        self.assertEqual(tags[0], "adam brian")
        self.assertEqual(tags[1], "chris")

    def test_order(self):
        tags = tag_utils.parse_tags("chris, adam, brian")
        self.assertEqual(len(tags), 3)
        self.assertEqual(tags[0], "adam")
        self.assertEqual(tags[1], "brian")
        self.assertEqual(tags[2], "chris")

    def test_quotes(self):
        tags = tag_utils.parse_tags('"adam, one"')
        self.assertEqual(len(tags), 1)
        self.assertEqual(tags[0], "adam, one")

    def test_quotes_comma_delim(self):
        tags = tag_utils.parse_tags('"adam, one","brian, two","chris, three"')
        self.assertEqual(len(tags), 3)
        self.assertEqual(tags[0], "adam, one")
        self.assertEqual(tags[1], "brian, two")
        self.assertEqual(tags[2], "chris, three")

    def test_quotes_space_delim(self):
        tags = tag_utils.parse_tags('"adam one" "brian two" "chris three"')
        self.assertEqual(len(tags), 3)
        self.assertEqual(tags[0], "adam one")
        self.assertEqual(tags[1], "brian two")
        self.assertEqual(tags[2], "chris three")

    def test_quotes_comma_delim_spaces_ignored(self):
        tags = tag_utils.parse_tags('"adam, one"  ,  "brian, two"  ,  "chris, three"')
        self.assertEqual(len(tags), 3)
        self.assertEqual(tags[0], "adam, one")
        self.assertEqual(tags[1], "brian, two")
        self.assertEqual(tags[2], "chris, three")

    def test_quotes_comma_delim_early_wins(self):
        tags = tag_utils.parse_tags('"adam one","brian two" "chris three"')
        self.assertEqual(len(tags), 2)
        self.assertEqual(tags[0], "adam one")
        self.assertEqual(tags[1], 'brian two" "chris three')

    def test_quotes_comma_delim_late_wins(self):
        tags = tag_utils.parse_tags('"adam one" "brian two","chris three"')
        self.assertEqual(len(tags), 2)
        self.assertEqual(tags[0], 'adam one" "brian two')
        self.assertEqual(tags[1], "chris three")

    def test_quotes_comma_delim_late_wins_unescaped_quotes(self):
        "When delimiter changes, return insignificant unescaped quotes"
        tags = tag_utils.parse_tags('adam "one", brian two')
        self.assertEqual(len(tags), 2)
        self.assertEqual(tags[0], 'adam "one"')
        self.assertEqual(tags[1], "brian two")

    def test_quotes_dont_delimit(self):
        tags = tag_utils.parse_tags('adam"brian,chris dave')
        self.assertEqual(len(tags), 2)
        self.assertEqual(tags[0], 'adam"brian')
        self.assertEqual(tags[1], "chris dave")

    def test_quotes_dont_close(self):
        tags = tag_utils.parse_tags('"adam,one","brian,two","chris, dave')
        self.assertEqual(len(tags), 3)
        # Will be sorted, " comes first
        self.assertEqual(tags[0], '"chris, dave')
        self.assertEqual(tags[1], "adam,one")
        self.assertEqual(tags[2], "brian,two")

    def test_quotes_and_unquoted(self):
        tags = tag_utils.parse_tags('adam , "brian, chris" , dave')
        self.assertEqual(len(tags), 3)
        self.assertEqual(tags[0], "adam")
        self.assertEqual(tags[1], "brian, chris")
        self.assertEqual(tags[2], "dave")

    def test_quotes_order(self):
        tags = tag_utils.parse_tags('chris, "adam", brian')
        self.assertEqual(len(tags), 3)
        self.assertEqual(tags[0], "adam")
        self.assertEqual(tags[1], "brian")
        self.assertEqual(tags[2], "chris")

    def test_quotes_escaped(self):
        """
        Tests quotes when delimiter is already comma
        """
        tags = tag_utils.parse_tags('adam, br""ian, ""chris, dave""')
        self.assertEqual(len(tags), 4)
        self.assertEqual(tags[0], '"chris')
        self.assertEqual(tags[1], "adam")
        self.assertEqual(tags[2], 'br"ian')
        self.assertEqual(tags[3], 'dave"')

    def test_quotes_escaped_late(self):
        """
        Tests quotes when delimiter switches to comma
        """
        tags = tag_utils.parse_tags('""adam"" brian"", chris')
        self.assertEqual(len(tags), 2)
        self.assertEqual(tags[0], '"adam" brian"')
        self.assertEqual(tags[1], "chris")

    def test_empty_tag(self):
        tags = tag_utils.parse_tags('"adam" , , brian , ')
        self.assertEqual(len(tags), 2)
        self.assertEqual(tags[0], "adam")
        self.assertEqual(tags[1], "brian")

    def test_limit(self):
        with self.assertRaises(ValueError) as cm:
            tag_utils.parse_tags("adam,brian,chris", 1)
        e = cm.exception
        self.assertEqual(str(e), "This field can only have 1 argument")

    def test_limit_quotes(self):
        with self.assertRaises(ValueError) as cm:
            tag_utils.parse_tags('"adam","brian",chris', 2)
        e = cm.exception
        self.assertEqual(str(e), "This field can only have 2 arguments")

    def test_spaces_false_commas(self):
        tags = tag_utils.parse_tags("adam,brian,chris", space_delimiter=False)
        self.assertEqual(len(tags), 3)
        self.assertEqual(tags[0], "adam")
        self.assertEqual(tags[1], "brian")
        self.assertEqual(tags[2], "chris")

    def test_spaces_false_spaces(self):
        tags = tag_utils.parse_tags("adam brian chris", space_delimiter=False)
        self.assertEqual(len(tags), 1)
        self.assertEqual(tags[0], "adam brian chris")

    def test_spaces_false_mixed(self):
        tags = tag_utils.parse_tags("adam,brian chris", space_delimiter=False)
        self.assertEqual(len(tags), 2)
        self.assertEqual(tags[0], "adam")
        self.assertEqual(tags[1], "brian chris")


# ##############################################################################
# ###### utils.render_tags
# ##############################################################################


class UtilsRenderTagsTest(TestCase):
    "Test utils.render_tags"

    def test_simple(self):
        tagstr = tag_utils.render_tags(["adam", "brian", "chris"])
        self.assertEqual(tagstr, "adam, brian, chris")

    def test_escapes_quotes(self):
        tagstr = tag_utils.render_tags(['ad"am', '"brian', 'chris"', '"dave"'])
        self.assertEqual(tagstr, '""brian, ""dave"", ad""am, chris""')

    def test_quotes_commas_and_spaces(self):
        tagstr = tag_utils.render_tags(["adam brian", "chris, dave", "ed"])
        self.assertEqual(tagstr, '"adam brian", "chris, dave", ed')

    def test_parse_renders_tags(self):
        tagstr = "adam, brian, chris"
        tags = tag_utils.parse_tags(tagstr)
        tagstr2 = tag_utils.render_tags(tags)
        self.assertEqual(len(tags), 3)
        self.assertEqual(tags[0], "adam")
        self.assertEqual(tags[1], "brian")
        self.assertEqual(tags[2], "chris")
        self.assertEqual(tagstr, tagstr2)

    def test_parse_renders_tags_complex(self):
        tagstr = '"""adam brian"", ""chris, dave", "ed, frank", gary'
        tags = tag_utils.parse_tags(tagstr)
        tagstr2 = tag_utils.render_tags(tags)
        self.assertEqual(len(tags), 3)
        self.assertEqual(tags[0], '"adam brian", "chris, dave')
        self.assertEqual(tags[1], "ed, frank")
        self.assertEqual(tags[2], "gary")
        self.assertEqual(tagstr, tagstr2)


# ##############################################################################
# ###### Tree name split and join
# ##############################################################################


class TagTreeSplitUtilTest(TestCase):
    """
    Test split_tree_name
    """

    def test_split_tree_none(self):
        parts = tag_utils.split_tree_name("")
        self.assertEqual(len(parts), 0)

    def test_split_tree_one(self):
        parts = tag_utils.split_tree_name("one")
        self.assertEqual(len(parts), 1)
        self.assertEqual(parts[0], "one")

    def test_split_tree_one_single_character(self):
        parts = tag_utils.split_tree_name("o")
        self.assertEqual(len(parts), 1)
        self.assertEqual(parts[0], "o")

    def test_split_tree_three(self):
        parts = tag_utils.split_tree_name("one/two/three")
        self.assertEqual(len(parts), 3)
        self.assertEqual(parts[0], "one")
        self.assertEqual(parts[1], "two")
        self.assertEqual(parts[2], "three")

    def test_split_tree_three_last_one_with_single_character(self):
        parts = tag_utils.split_tree_name("one/two/3")
        self.assertEqual(len(parts), 3)
        self.assertEqual(parts[0], "one")
        self.assertEqual(parts[1], "two")
        self.assertEqual(parts[2], "3")

    def test_split_tree_three_spaced(self):
        parts = tag_utils.split_tree_name("  one  /  two  /  three  ")
        self.assertEqual(len(parts), 3)
        self.assertEqual(parts[0], "one")
        self.assertEqual(parts[1], "two")
        self.assertEqual(parts[2], "three")

    def test_split_tree_leading(self):
        parts = tag_utils.split_tree_name("/one")
        self.assertEqual(len(parts), 2)
        self.assertEqual(parts[0], "")
        self.assertEqual(parts[1], "one")

    def test_split_tree_trailing(self):
        parts = tag_utils.split_tree_name("one/")
        self.assertEqual(len(parts), 2)
        self.assertEqual(parts[0], "one")
        self.assertEqual(parts[1], "")

    def test_split_tree_escape(self):
        parts = tag_utils.split_tree_name("one/two//dos/three")
        self.assertEqual(len(parts), 3)
        self.assertEqual(parts[0], "one")
        self.assertEqual(parts[1], "two/dos")
        self.assertEqual(parts[2], "three")

    def test_split_tree_escape_odd(self):
        parts = tag_utils.split_tree_name("one/two///three/four")
        self.assertEqual(len(parts), 4)
        self.assertEqual(parts[0], "one")
        self.assertEqual(parts[1], "two/")
        self.assertEqual(parts[2], "three")
        self.assertEqual(parts[3], "four")

    def test_split_tree_escape_even(self):
        parts = tag_utils.split_tree_name("one/two////dos/three")
        self.assertEqual(len(parts), 3)
        self.assertEqual(parts[0], "one")
        self.assertEqual(parts[1], "two//dos")
        self.assertEqual(parts[2], "three")

    def test_split_tree_escape_leading(self):
        parts = tag_utils.split_tree_name("//one/two")
        self.assertEqual(len(parts), 2)
        self.assertEqual(parts[0], "/one")
        self.assertEqual(parts[1], "two")

    def test_split_tree_escape_trailing(self):
        parts = tag_utils.split_tree_name("one/two//")
        self.assertEqual(len(parts), 2)
        self.assertEqual(parts[0], "one")
        self.assertEqual(parts[1], "two/")


class TagTreeJoinUtilTest(TestCase):
    """
    Test join_tree_name
    """

    def test_join_tree_none(self):
        name = tag_utils.join_tree_name([])
        self.assertEqual(name, "")

    def test_join_tree_one(self):
        name = tag_utils.join_tree_name(["one"])
        self.assertEqual(name, "one")

    def test_join_tree_three(self):
        name = tag_utils.join_tree_name(["one", "two", "three"])
        self.assertEqual(name, "one/two/three")

    def test_join_tree_escape(self):
        name = tag_utils.join_tree_name(["one", "two/dos", "three"])
        self.assertEqual(name, "one/two//dos/three")

    def test_join_tree_escape_odd(self):
        name = tag_utils.join_tree_name(["one/", "two"])
        self.assertEqual(name, "one///two")

    def test_join_tree_escape_even(self):
        name = tag_utils.join_tree_name(["one", "two//dos", "three"])
        self.assertEqual(name, "one/two////dos/three")

    def test_join_tree_escape_leading(self):
        name = tag_utils.join_tree_name(["/one", "two"])
        self.assertEqual(name, "//one/two")

    def test_join_tree_escape_trailing(self):
        name = tag_utils.join_tree_name(["one", "two/"])
        self.assertEqual(name, "one/two//")


class TagTreeCleanUtilTest(TestCase):
    """
    Test clean_tree_name
    """

    def test_clean_tree_one(self):
        name = tag_utils.clean_tree_name("one")
        self.assertEqual(name, "one")

    def test_clean_tree_three(self):
        name = tag_utils.clean_tree_name("one/two/three")
        self.assertEqual(name, "one/two/three")

    def test_clean_tree_escape(self):
        name = tag_utils.clean_tree_name("one/two//dos/three")
        self.assertEqual(name, "one/two//dos/three")

    def test_clean_tree_strip(self):
        name = tag_utils.clean_tree_name("  one  /  two  /  three  ")
        self.assertEqual(name, "one/two/three")

    def test_clean_tree_leading(self):
        name = tag_utils.clean_tree_name("/one/two/three")
        self.assertEqual(name, "//one/two/three")

    def test_clean_tree_trailing(self):
        name = tag_utils.clean_tree_name("one/two/three//")
        self.assertEqual(name, "one/two/three//")

    def test_clean_tree_complex(self):
        name = tag_utils.clean_tree_name("/// one / two/ three /")
        self.assertEqual(name, "//// one/two/three //")



from django.contrib import admin

from tests.tagulous_tests_app import models


class SimpleMixedTestAdmin(admin.ModelAdmin):
    """
    For testing tagged models
    For use with models.SimpleMixedTest
    """

    list_display = ("name", "singletag", "tags")
    list_filter = ["singletag", "tags"]
    fields = ("name", "singletag", "tags")
    list_display_links = [None]


class SimpleMixedTestInline(admin.TabularInline):
    """
    For testing tagged model inlines on tag model admin
    """

    model = models.SimpleMixedTest
    extra = 3


class SimpleMixedTestSingletagAdmin(admin.ModelAdmin):
    "For testing FK inlines"
    inlines = [SimpleMixedTestInline]


class SimpleMixedTestTagsAdmin(admin.ModelAdmin):
    """
    For testing attribute overrides on TagModels
    For use with models.SimpleMixedTest.tags.tag_model
    """

    list_display = ["name"]
    list_filter = ["count"]
    exclude = ["name"]
    actions = []
    inlines = [SimpleMixedTestInline]



# -*- coding: utf-8 -*-
"""
Django Tagulous - Fabulous Tags
"""
__version__ = "1.2.1"
__license__ = "BSD"
__author__ = "Richard Terry"
__credits__ = [
    "Dmitry Ivanchenko (ivanchenkodmitry)",
    "Erik Van Kelst (4levels)",
    "Frank Lanitz (frlan)",
    "Gaël Utard (gutard)",
    "James Pic (jpic)",
    "Khoa Pham (phamk)",
    "Mark London",
    "Martín R. Guerrero (slackmart)",
    "Mary (minidietcoke)",
    "Michael Röttger (mcrot)",
    "Pamela McA'Nulty (PamelaM)",
    "Peter Baumgartner (ipmb)",
    "Raniere Silva (rgaics)",
    "Robert Erb (rerb) at AASHE",
    "hagsteel",
    "Diego Ubirajara (dubirajara)",
    "Andrew O'Brien (marxide)",
]

default_app_config = "tagulous.apps.TagulousConfig"



"""
Migration support for Django migrations
"""
from django.db import migrations

from .models import BaseTagModel, BaseTagTreeModel


# ##############################################################################
# ########################################################### Django migrations
# ##############################################################################


def django_support():
    # Monkey-patch Django so it doesn't flatten TagModel abstract base classes.
    # We need them so the BaseTagModel metaclass can set up its options, and
    # so that tag model functionality is available in migrations.
    old_from_model = migrations.state.ModelState.from_model.__func__

    def from_model(cls, model, exclude_rels=False):
        base = None
        if issubclass(model, BaseTagTreeModel):
            base = BaseTagTreeModel
        elif issubclass(model, BaseTagModel):
            base = BaseTagModel

        modelstate = old_from_model(cls, model, exclude_rels)

        if base is not None:
            modelstate.bases = (base,) + modelstate.bases
        return modelstate

    migrations.state.ModelState.from_model = classmethod(from_model)


# Add Tagulous support to Django migrations
django_support()


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#       Helpers for Django migrations
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class ChangeModelBases(migrations.operations.base.Operation):
    """
    Change a Model's bases so it is correct for data migrations
    """

    def __init__(self, name, bases):
        self.name = name
        self.bases = bases

    def state_forwards(self, app_label, state):
        state.models[app_label, self.name.lower()].bases = self.bases

    def database_forwards(self, *args, **kwargs):
        pass

    def database_backwards(self, *args, **kwargs):
        pass


def add_unique_field(model_name, name, field, preserve_default, set_fn):
    """
    Helper for Django migrations which returns a list of Operations to add a
    unique field.

    Warning: only use on a database which supports transactions, in case your
    set_fn method fails and leaves your database in an unusable state.

    Create a migration and provide the string 'x' as the default when prompted.

    Then find the field definition in your migration and replace it with a call
    to this function, removing the default:

        operations = [
            migrations.AddField(
                ...
            ),
            migrations.AddField(
                model_name='mymodel',
                name='my_column',
                field=models.TextField(default='x', unique=True),
                preserve_default=False,
            ),
            migrations.AddField(
                ...
            ),
        ]

    becomes:

        def set_new_column(obj):
            obj.new_column = slugify(obj.name)

        operations = [
            migrations.AddField(
                ...
            ),
        ] + tagulous.models.migration.add_unique_field(
            model_name='mymodel',
            name='my_column',
            field=models.TextField(unique=True),
            preserve_default=False,
            set_fn=set_new_column,
        ) + [
            migrations.AddField(
                ...
            ),
        ]

    Arguments:
        model_name  As django defines
        name        As django defines
        field       As django defines, but without default
        preserve_default    As django defines
        set_fn      Callback to set the field on each instance
    """
    if migrations is None:  # pragma: no cover
        raise ValueError("Cannot use add_unique_column without Django migrations")

    # Clone the field so it's not unique and can be null
    field_nullable = field.clone()
    field_nullable._unique = False
    field_nullable.null = True

    # RunPython doesn't give the code the app label; pass it as an attribute
    # to save users having to specify the app
    class RunPythonWithAppLabel(migrations.RunPython):
        def database_forwards(self, app_label, *args, **kwargs):
            self.code.app_label = app_label
            super(RunPythonWithAppLabel, self).database_forwards(
                app_label, *args, **kwargs
            )

        def database_backwards(self, app_label, *args, **kwargs):
            self.code.app_label = app_label
            super(RunPythonWithAppLabel, self).database_backwards(
                app_label, *args, **kwargs
            )

    # Function to generate the unique value
    def set_unique_values(apps, schema_editor):
        model = apps.get_model(set_unique_values.app_label, model_name)

        # Make sure tag models won't mess with data during this operation
        is_tag_model = issubclass(model, BaseTagModel)
        for obj in model.objects.all():
            set_fn(obj)
            if is_tag_model:
                obj._save_direct()
            else:  # pragma: no cover - no need, _save_direct() calls save()
                obj.save()

    return [
        migrations.AddField(
            model_name=model_name,
            name=name,
            field=field_nullable,
            preserve_default=preserve_default,
        ),
        RunPythonWithAppLabel(set_unique_values, reverse_code=lambda *a, **k: None),
        migrations.AlterField(
            model_name=model_name,
            name=name,
            field=field,
            preserve_default=preserve_default,
        ),
    ]



"""
YAML serializer with Tagulous support
"""
from django.core.serializers import pyyaml as pyyaml_serializer

from . import base


class Serializer(base.SerializerMixin, pyyaml_serializer.Serializer):
    """
    YAML serializer with tag field support
    """

    pass


Deserializer = base.DeserializerWrapper(
    pyyaml_serializer.Deserializer,
    doc="Deserialize a stream or string of YAML data, with tag field support",
)



"""
Tagulous test of serializers, dumpdata and loaddata

Modules tested:
    tagulous.serializers.*
"""
# There were originally separate tests for loaddata and dumpdata, but these were
# combined because in Django <1.8 serializer order was not deterministic.
#
# This has been addressed, so we can look at splitting them again when these tests need
# refactoring
#
import os
import re
import tempfile
import unittest
from io import StringIO

from django.core import management, serializers
from django.test import TestCase

from tests.lib import TagTestManager, testenv
from tests.tagulous_tests_app import models as test_models


try:
    import yaml
except ImportError:
    yaml = None


RE_STRIP_PK_JSON = re.compile(r'("(pk|fk)": \d+|"mm": \[[0-9, ]*\])')
RE_STRIP_PK_YAML = re.compile(r"((pk|fk): \d+|mm:\s*(\n\s*- \d+)+)")
RE_STRIP_PK_XML = re.compile(r'(pk="\d+"|<field name="fk"[^>]+>\d+</field>)')

# Based on django's tests/fixtures/tests.py -> DumpDataAssertMixin
# but without comparisons due to django#24558
class DumpDataAssertMixin(object):
    def _dumpdata(
        self,
        args,
        format="json",
        filename=None,
        natural_foreign_keys=False,
        natural_primary_keys=False,
        use_base_manager=False,
        exclude_list=[],
        primary_keys="",
    ):
        new_io = StringIO()
        if filename:
            filename = os.path.join(tempfile.gettempdir(), filename)
        management.call_command(
            "dumpdata",
            *args,
            **{
                "format": format,
                "stdout": new_io,
                "stderr": new_io,
                "output": filename,
                "use_natural_foreign_keys": natural_foreign_keys,
                "use_natural_primary_keys": natural_primary_keys,
                "use_base_manager": use_base_manager,
                "exclude": exclude_list,
                "primary_keys": primary_keys,
            },
        )
        if filename:
            with open(filename, "r") as f:
                command_output = f.read()
            os.remove(filename)
        else:
            command_output = new_io.getvalue().strip()

        # Return early, do comparison outside this mixin
        return command_output


fixture_root = os.path.join(os.path.dirname(__file__), "tagulous_tests_app", "fixtures")


# ##############################################################################
# ###### Test serialization of a tagged model
# ##############################################################################


class SerializationTestMixin(DumpDataAssertMixin):
    """
    Test JSON serializer
    """

    manage_models = [test_models.SimpleMixedTest]
    dump_models = [
        "tagulous_tests_app.Tagulous_SimpleMixedTest_singletag",
        "tagulous_tests_app.Tagulous_SimpleMixedTest_tags",
        "tagulous_tests_app.SimpleMixedTest",
    ]
    fixture_format = "undefined"
    fixture_prefix = "simple"

    def setUpExtra(self):
        self.model = test_models.SimpleMixedTest
        self.singletag_model = self.model.singletag.tag_model
        self.tags_model = self.model.tags.tag_model
        self.fixture_name = "test_fixtures.%s" % self.fixture_format
        self.fixture_path = os.path.join(
            fixture_root, "{}_{}".format(self.fixture_prefix, self.fixture_name)
        )

        # Need to dump to loadable file due to django#24558
        self.tmp_fixture_name = "tmp_%s_%s" % (testenv, self.fixture_name)
        self.tmp_fixture_path = os.path.join(fixture_root, self.tmp_fixture_name)

        # Delete file if it exists
        if os.path.exists(self.tmp_fixture_path):
            os.remove(self.tmp_fixture_path)

    def tearDownExtra(self):
        if os.path.exists(self.tmp_fixture_path):
            os.remove(self.tmp_fixture_path)

    def _populate(self):
        self.t1 = self.model.objects.create(
            name="Test 1", singletag="single1", tags="tag1, tag2"
        )
        self.t2 = self.model.objects.create(
            name="Test 2", singletag="single2", tags="tag3, tag4"
        )

    def _check(self):
        t1 = self.model.objects.get(name="Test 1")
        t2 = self.model.objects.get(name="Test 2")
        self.assertInstanceEqual(
            t1, name="Test 1", singletag="single1", tags="tag1, tag2"
        )
        self.assertInstanceEqual(
            t2, name="Test 2", singletag="single2", tags="tag3, tag4"
        )
        self.assertEqual(t1.singletag.name, "single1")
        self.assertEqual(t2.singletag.name, "single2")
        self.assertEqual(t1.tags.count(), 2)
        self.assertEqual(t2.tags.count(), 2)
        self.assertTagModel(self.singletag_model, {"single1": 1, "single2": 1})
        self.assertTagModel(
            self.tags_model, {"tag1": 1, "tag2": 1, "tag3": 1, "tag4": 1}
        )

    def _assert_dumped(self, dumped):
        with open(self.fixture_path) as file:
            raw = file.read()
        self.assert_equivalent(raw, dumped)

    def assert_equivalent(self, left, right):
        """
        Assert two fixtures are equivalent, allowing for pk variation
        """
        left = self.re_strip_pk.sub("", left)
        right = self.re_strip_pk.sub("", right)
        self.assertEqual(left, right)

    def _empty(self):
        self.t1.delete()
        self.t2.delete()
        self.assertEqual(self.model.objects.count(), 0)
        self.assertTagModel(self.singletag_model, {})
        self.assertTagModel(self.tags_model, {})

    def test_dump_load(self):
        "Check dumpdata and loaddata recreate data"
        # Populate and confirm
        self._populate()
        self._check()

        # Dump
        dumped = self._dumpdata(self.dump_models, format=self.fixture_format)
        self._assert_dumped(dumped)

        # Fixture path
        with open(self.tmp_fixture_path, "w") as f:
            f.write(dumped)

        # Empty the database
        self._empty()

        # Now load fixture
        management.call_command("loaddata", self.tmp_fixture_name, verbosity=0)

        # Check they were loaded correctly
        self._check()


class JsonSerializationTest(SerializationTestMixin, TagTestManager, TestCase):
    "Test JSON serializer"
    fixture_format = "json"
    re_strip_pk = RE_STRIP_PK_JSON


@unittest.skipIf(yaml is None, "pyyaml is not installed")
class YamlSerializationTest(SerializationTestMixin, TagTestManager, TestCase):
    """
    Test YAML serializer
    """

    fixture_format = "yaml"
    re_strip_pk = RE_STRIP_PK_YAML


class XmlSerializationTest(SerializationTestMixin, TagTestManager, TestCase):
    """
    Test XML serializer
    """

    fixture_format = "xml"
    re_strip_pk = RE_STRIP_PK_XML


# ##############################################################################
# ###### Check serialization of tagged model with conventional FK and M2M fields
# ##############################################################################


class MixedTestMixin(SerializationTestMixin):
    manage_models = [test_models.MixedNonTagRefTest]
    dump_models = [
        "tagulous_tests_app.MixedNonTagModel",
        "tagulous_tests_app.MixedNonTagRefTest",
    ]
    fixture_prefix = "mixed"

    def setUpExtra(self):
        self.model = test_models.MixedNonTagRefTest
        self.tag_model = test_models.MixedNonTagModel
        self.fixture_name = "test_fixtures.%s" % self.fixture_format
        self.fixture_path = os.path.join(
            fixture_root, "{}_{}".format(self.fixture_prefix, self.fixture_name)
        )

        # Need to dump to loadable file due to django#24558
        self.tmp_fixture_name = "tmp_%s_%s" % (testenv, self.fixture_name)
        self.tmp_fixture_path = os.path.join(fixture_root, self.tmp_fixture_name)

        # Delete file if it exists
        if os.path.exists(self.tmp_fixture_path):
            os.remove(self.tmp_fixture_path)

    def _populate(self):
        tag1 = self.tag_model.objects.create(name="tag1")
        tag2 = self.tag_model.objects.create(name="tag2")
        tag3 = self.tag_model.objects.create(name="tag3")
        tag4 = self.tag_model.objects.create(name="tag4")
        self.t1 = self.model.objects.create(
            name="Test 1", singletag=tag1, tags=[tag1, tag2], fk=tag1
        )
        self.t1.mm.add(tag1, tag2)
        self.t2 = self.model.objects.create(
            name="Test 2", singletag=tag2, tags=[tag3, tag4], fk=tag2
        )
        self.t2.mm.add(tag3, tag4)

    def _check(self):
        t1 = self.model.objects.get(name="Test 1")
        t2 = self.model.objects.get(name="Test 2")
        tag1 = self.tag_model.objects.get(name="tag1")
        tag2 = self.tag_model.objects.get(name="tag2")
        tag3 = self.tag_model.objects.get(name="tag3")
        tag4 = self.tag_model.objects.get(name="tag4")
        self.assertInstanceEqual(
            t1,
            name="Test 1",
            singletag="tag1",
            tags="tag1, tag2",
            fk="tag1",
            mm=[tag1, tag2],
        )
        self.assertInstanceEqual(
            t2,
            name="Test 2",
            singletag="tag2",
            tags="tag3, tag4",
            fk="tag2",
            mm=[tag3, tag4],
        )
        self.assertTagModel(
            self.tag_model, {"tag1": 2, "tag2": 2, "tag3": 1, "tag4": 1}
        )

    def _empty(self):
        self.t1.delete()
        self.t2.delete()
        self.tag_model.objects.all().delete()
        self.assertEqual(self.model.objects.count(), 0)
        self.assertTagModel(self.tag_model, {})


class MixedJsonSerializationTest(MixedTestMixin, TagTestManager, TestCase):
    "Test JSON serializer with normal fk and m2m fields"
    fixture_format = "json"
    re_strip_pk = RE_STRIP_PK_JSON


@unittest.skipIf(yaml is None, "pyyaml is not installed")
class MixedYamlSerializationTest(MixedTestMixin, TagTestManager, TestCase):
    "Test Yaml serializer with normal fk and m2m fields"
    fixture_format = "yaml"
    re_strip_pk = RE_STRIP_PK_YAML


class MixedXmlSerializationTest(MixedTestMixin, TagTestManager, TestCase):
    "Test XML serializer with normal fk and m2m fields"
    fixture_format = "xml"
    re_strip_pk = RE_STRIP_PK_XML


# ##############################################################################
# ###### Other serialization tests
# ##############################################################################


class MixedTestMixin(TagTestManager, TestCase):
    def test_many_to_one(self):
        """
        Check serialization of a tagged model with reverse FKs
        """
        t1 = test_models.MixedRefTest.objects.create(
            name="test", singletag="test", tags="test"
        )
        rfk1 = test_models.ManyToOneTest.objects.create(name="rfk1", mixed_ref_test=t1)
        t1.refresh_from_db()
        self.assertEqual(t1.many_to_one.count(), 1)
        self.assertEqual(t1.many_to_one.first(), rfk1)

        serialized = serializers.serialize(
            "xml", test_models.MixedRefTest.objects.all()
        )
        deserialized = list(serializers.deserialize("xml", serialized))
        self.assertEqual(len(deserialized), 1)
        obj = deserialized[0].object
        self.assertInstanceEqual(obj, name="test", singletag="test", tags="test")
        self.assertEqual(obj.many_to_one.count(), 1)
        self.assertEqual(obj.many_to_one.first().name, "rfk1")



"""
Tagulous model field descriptors

When a model has a SingleTagField or TagField, they are replaced with these
descriptors during the ``contribute_to_class`` phase.

Their main purposes is to act as getter/setters and pass data to and from
manager instances.
"""
from collections.abc import Iterable

from .managers import FakeTagRelatedManager, SingleTagManager, TagRelatedManagerMixin


# ##############################################################################
# ###### Base class for tag field descriptors
# ##############################################################################


class BaseTagDescriptor(object):
    """
    Base TagDescriptor class
    """

    def __init__(self, descriptor):
        # Store original FK/M2M descriptor and tag options
        self.descriptor = descriptor

        # Copy descriptor attributes
        for key, val in descriptor.__dict__.items():
            setattr(self, key, val)

    # If the field was created using a string, the field's tag model and
    # tag options will not be available when the descriptor is created, so
    # cannot store them directly here
    tag_model = property(lambda self: self.field.remote_field.model)
    tag_options = property(lambda self: self.field.tag_options)

    def load_initial(self):
        """
        Load initial tags
        Be prepared to receive a DatabaseError if the model has not been synced
        """
        for tag_name in self.tag_options.initial:
            self.tag_model.objects.get_or_create(
                name=tag_name, defaults={"protected": self.tag_options.protect_initial}
            )

    def formfield(self, *args, **kwargs):
        """
        Shortcut to access formfield
        """
        return self.descriptor.field.formfield(*args, **kwargs)


# ##############################################################################
# ###### Descriptor for SingleTagField
# ##############################################################################


class SingleTagDescriptor(BaseTagDescriptor):
    """
    Descriptor to set the tag string rather than a Tag object
    Wraps the ReverseSingleRelatedObjectDescriptor and passes set and get
    requests through to the SingleTagManager
    """

    def __set__(self, instance, value):
        # Check we've actually got an instance. No practical way this could
        # happen, but Django does it, so we will too
        if instance is None:  # pragma: no cover
            raise AttributeError("Manager must be accessed via instance")

        # Otherwise set on the manager
        manager = self.get_manager(instance)
        manager.set(value)

    def get_manager(self, instance, instance_type=None):
        """
        Get the Manager instance for this field on this model instance.

        Descriptor is instantiated once per class, so we bind the Manager to
        the instance and collect it again when needed.
        """
        # Get existing or create new SingleTagManager
        attname = self.descriptor.field.get_manager_name()
        manager = getattr(instance, attname, None)
        if not manager:
            manager = SingleTagManager(self, instance)
            setattr(instance, attname, manager)
        return manager

    def __get__(self, instance, instance_type=None):
        # If no instance, return self
        if not instance:
            return self

        # Otherwise get from the manager
        manager = self.get_manager(instance, instance_type)
        return manager.get()


# ##############################################################################
# ###### Descriptor for TagField
# ##############################################################################


class TagDescriptor(BaseTagDescriptor):
    """
    Descriptor to add tag functions to the RelatedManager
    The ManyToManyField will create a ReverseManyRelatedObjectsDescriptor
    This will use a RelatedManager which we cannot customise
    This will intercept calls for the RelatedManager, and add the tag functions
    """

    def get_manager(self, instance, instance_type=None):
        """
        Get the Manager instance for this field on this model instance.

        Descriptor is instantiated once per class, so we bind the Manager to
        the instance and collect it again when needed.

        While the instance is not saved (does not have a pk), a fake manager
        is used to hold unsaved tags; database methods are disallowed.
        """
        # Get existing mamager
        attname = self.descriptor.field.get_manager_name()
        manager = getattr(instance, attname, None)
        manager_type = self.descriptor.related_manager_cls

        # Find which manager we should be using
        if instance.pk:
            # Can use real manager
            if isinstance(manager, FakeTagRelatedManager):
                # Convert fake manager to real manager
                fake_manager = manager
                manager = self.create_manager(instance, instance_type)
                manager.load_from_tagmanager(fake_manager)

            elif manager is None:
                # Create real manager
                manager = self.create_manager(instance, instance_type)

            # Otherwise already have real manager
        else:
            # Have to use the fake one
            if isinstance(manager, manager_type):
                # Was using a real manager, but instance must have been deleted
                real_manager = manager
                manager = FakeTagRelatedManager(self, instance, instance_type)
                manager.load_from_tagmanager(real_manager)

            elif manager is None:
                # Create fake manager
                manager = FakeTagRelatedManager(self, instance, instance_type)

            # Otherwise already have a fake manager

        # Set it in case it changed
        setattr(instance, attname, manager)
        return manager

    def create_manager(self, instance, instance_type):
        # Get the RelatedManager that should have been returned
        manager = self.descriptor.__get__(instance, instance_type)

        # Add in the mixin
        manager.__class__ = type(
            str("TagRelatedManager"), (TagRelatedManagerMixin, manager.__class__), {}
        )

        # Manager is already instantiated; initialise tagulous in it
        manager.init_tagulous(self)

        return manager

    def __get__(self, instance, instance_type=None):
        # If no instance, return self
        if not instance:
            return self

        # Otherwise get from the manager
        manager = self.get_manager(instance, instance_type)
        return manager

    def __set__(self, instance, value):
        # Check we've actually got an instance. No practical way this could
        # happen, but Django does it, so we will too
        if instance is None:  # pragma: no cover
            raise AttributeError("Manager must be accessed via instance")

        # Get the manager
        manager = self.__get__(instance)

        # Set value
        if not value:
            # Clear
            manager.set_tag_string("")

        elif isinstance(value, str):
            # If it's a string, it must be a tag string
            manager.set_tag_string(value)

        elif isinstance(value, Iterable):
            # An iterable goes in as a list of things that are, or can be
            # converted to, strings
            manager.set_tag_list(value)

        else:
            # Unknown
            raise ValueError("Unexpected value assigned to TagField")

    @property
    def through(self):
        return self.descriptor.through



"""
Tagulous test: Enhanced queryset for tagged models

Will fail if settings.ENHANCE_MODELS is not True

Modules tested:
    tagulous.models.tagged
"""
import pickle

from django.core.exceptions import MultipleObjectsReturned
from django.db import models
from django.test import TestCase

from tagulous import models as tag_models
from tagulous.models.tagged import _split_kwargs
from tests.lib import TagTestManager, skip_if_mysql
from tests.tagulous_tests_app import models as test_models


# ##############################################################################
# ###### Test TaggedModel
# ##############################################################################


class ModelTaggedTest(TagTestManager, TestCase):
    """
    Basic TaggedModel tests
    """

    manage_models = [test_models.MixedTest]

    def setUpExtra(self):
        self.test_model = test_models.MixedTest

    def test_tagged_model(self):
        "Check tagged model is the correct class"
        self.assertTrue(issubclass(self.test_model, tag_models.TaggedModel))

    def test_tagged_manager(self):
        "Check tagged manager is the correct class"
        self.assertTrue(
            issubclass(self.test_model.objects.__class__, tag_models.TaggedManager)
        )

    def test_base_manager_not_tagged(self):
        "Check normal base manager hasn't been modified"
        self.assertFalse(issubclass(models.Manager, tag_models.TaggedManager))

    def test_init(self):
        "Check constructor can set singletag and tag fields"
        t1 = self.test_model(name="Test 1", singletag="Mr", tags="red, blue")
        t1.save()
        t2 = self.test_model.objects.get(name="Test 1")
        self.assertEqual(str(t2.singletag), "Mr")
        self.assertEqual(str(t2.tags), "blue, red")

    def test_pickle(self):
        "Check tagged models can be pickled"
        t1 = self.test_model(name="Test 1", singletag="Mr", tags="red, blue")
        t1.save()
        pickled = pickle.dumps(t1)
        t2 = pickle.loads(pickled)
        self.assertEqual(str(t2.singletag), "Mr")
        self.assertEqual(str(t2.tags), "blue, red")


# ##############################################################################
# ###### Test _split_kwargs
# ##############################################################################


class SplitKwargsText(TagTestManager, TestCase):
    "Test _split_kwargs"

    def setUpExtra(self):
        self.test_model = test_models.MixedTest

    def test_fields(self):
        "Look up known fields"
        safe_fields, singletag_fields, tag_fields = _split_kwargs(
            self.test_model, {"name": "Adam", "singletag": "Mr", "tags": "red, blue"}
        )
        self.assertEqual(len(safe_fields), 1)
        self.assertEqual(safe_fields["name"], "Adam")

        self.assertEqual(len(singletag_fields), 1)
        self.assertEqual(singletag_fields["singletag"], "Mr")

        self.assertEqual(len(tag_fields), 1)
        self.assertEqual(tag_fields["tags"], "red, blue")

    def test_with_fields(self):
        "Fields dict returned when with_fields"
        safe_fields, singletag_fields, tag_fields, field_lookup = _split_kwargs(
            self.test_model,
            {"name": "Adam", "singletag": "Mr", "tags": "red, blue"},
            with_fields=True,
        )
        self.assertEqual(len(safe_fields), 1)
        self.assertEqual(len(singletag_fields), 1)
        self.assertEqual(len(tag_fields), 1)

        self.assertEqual(len(field_lookup), 3)
        self.assertEqual(
            field_lookup["name"], test_models.MixedTest._meta.get_field("name")
        )
        self.assertEqual(
            field_lookup["singletag"],
            test_models.MixedTest._meta.get_field("singletag"),
        )
        self.assertEqual(
            field_lookup["tags"], test_models.MixedTest._meta.get_field("tags")
        )

    def test_unknown(self):
        "Unknown fields passed as safe"
        safe_fields, singletag_fields, tag_fields = _split_kwargs(
            self.test_model, {"unknown": "field"}
        )
        self.assertEqual(len(safe_fields), 1)
        self.assertEqual(safe_fields["unknown"], "field")
        self.assertEqual(len(singletag_fields), 0)
        self.assertEqual(len(tag_fields), 0)

    def test_lookups(self):
        "Test with lookups enabled"
        safe_fields, singletag_fields, tag_fields = _split_kwargs(
            self.test_model,
            {
                "name__startswith": "Ad",
                "singletag__name__contains": "r",
                "tags__name__contains": "e",
                "tags__exact": "red, blue",
            },
            lookups=True,
        )

        # Unknown lookups are passed straight
        self.assertEqual(len(safe_fields), 3)
        self.assertEqual(safe_fields["name__startswith"], "Ad")
        self.assertEqual(safe_fields["singletag__name__contains"], "r")
        self.assertEqual(safe_fields["tags__name__contains"], "e")

        self.assertEqual(len(singletag_fields), 0)

        self.assertEqual(len(tag_fields), 1)
        self.assertEqual(tag_fields["tags"], ("red, blue", "exact"))

    def test_lookups_normal(self):
        "Test with lookups enabled but normal tag"
        safe_fields, singletag_fields, tag_fields = _split_kwargs(
            self.test_model, {"tags": "red"}, lookups=True
        )

        # Unknown lookups are passed straight
        self.assertEqual(len(safe_fields), 0)
        self.assertEqual(len(singletag_fields), 0)

        self.assertEqual(len(tag_fields), 1)
        self.assertEqual(tag_fields["tags"], ("red", None))

    def test_lookups_unknown(self):
        "Unknown fields in lookups passed as safe"
        safe_fields, singletag_fields, tag_fields = _split_kwargs(
            self.test_model, {"unknown__exact": "field"}, lookups=True
        )
        self.assertEqual(len(safe_fields), 1)
        self.assertEqual(safe_fields["unknown__exact"], "field")
        self.assertEqual(len(singletag_fields), 0)
        self.assertEqual(len(tag_fields), 0)


# ##############################################################################
# ###### Test TaggedManager and TaggedQuerySet
# ##############################################################################


class ModelTaggedQuerysetTest(TagTestManager, TestCase):
    """
    Test enhanced tagged model querysets
    """

    manage_models = [test_models.MixedTest]

    def setUpExtra(self):
        self.test_model = test_models.MixedTest

        self.o1 = self.test_model.objects.create(
            name="Test 1", singletag="Mr", tags="red, green, blue"
        )
        self.o2 = self.test_model.objects.create(
            name="Test 2", singletag="Mrs", tags="red, green, blue"
        )
        self.o3 = self.test_model.objects.create(
            name="Test 3", singletag="Mr", tags="red, green"
        )

    #
    # .get()
    #

    def test_object_get_by_singletag(self):
        "Check that object.get singletag loads the item correctly"
        t1 = self.test_model.objects.get(singletag="Mrs")
        self.assertEqual(t1.pk, self.o2.pk)

    def test_object_get_by_singletag_multiple_raises(self):
        "Check that object.get singletag raises exception when multiple found"
        with self.assertRaises(MultipleObjectsReturned) as cm:
            self.test_model.objects.get(singletag="Mr")
        self.assertTrue(
            str(cm.exception).startswith(
                "get() returned more than one MixedTest -- it returned 2!"
            )
        )

    def test_object_get_by_tags(self):
        "Check that object.get tags loads the item correctly"
        o4 = self.test_model.objects.create(name="Test 4", tags="red, green, yellow")
        t1 = self.test_model.objects.get(tags="red, yellow")
        self.assertEqual(t1.pk, o4.pk)

    def test_object_get_by_tags_multiple_raises(self):
        "Check that object.get tags raises exception when multiple found"
        with self.assertRaises(MultipleObjectsReturned) as cm:
            self.test_model.objects.get(tags="red, green")
        self.assertTrue(
            str(cm.exception).startswith(
                "get() returned more than one MixedTest -- it returned 3!"
            )
        )

    def test_object_get_by_tags_exact(self):
        "Check that object.get tags loads the item correctly"
        t1 = self.test_model.objects.get(tags__exact="red, green")
        self.assertEqual(t1.pk, self.o3.pk)

    def test_object_get_by_tags_multiple_exact_raises(self):
        "Check that object.get tags raises exception when multiple found"
        self.test_model.objects.create(name="Test 4", tags="yellow, purple")
        self.test_model.objects.create(name="Test 5", tags="yellow, purple")
        with self.assertRaises(MultipleObjectsReturned) as cm:
            self.test_model.objects.get(tags__exact="yellow, purple")
        self.assertTrue(
            str(cm.exception).startswith(
                "get() returned more than one MixedTest -- it returned 2!"
            )
        )

    def test_object_get_by_singletag_tags(self):
        "Check that object.get singletag and tags loads the item correctly"
        t1 = self.test_model.objects.get(singletag="Mr", tags="red, blue")
        self.assertEqual(t1.pk, self.o1.pk)

    #
    # .get_or_create()
    #

    def test_object_get_or_create_singletagfield(self):
        "Check that objects.get_or_create on SingleTagField creates and gets"
        # Clear out models
        self.test_model.objects.all().delete()
        self.test_model.singletag.tag_model.objects.all().delete()
        self.assertTagModel(self.test_model.singletag, {})

        # Create
        t1, created = self.test_model.objects.get_or_create(singletag="Mr")
        self.assertEqual(created, True)
        self.assertTagModel(self.test_model.singletag, {"Mr": 1})
        t1.name = "Test 4"
        t1.save()
        self.assertEqual(str(t1.singletag), "Mr")

        # Get
        t2, created = self.test_model.objects.get_or_create(singletag="Mr")
        self.assertEqual(created, False)
        self.assertTagModel(self.test_model.singletag, {"Mr": 1})
        self.assertEqual(t2.name, "Test 4")
        self.assertEqual(str(t2.singletag), "Mr")

        self.assertEqual(t1.pk, t2.pk)

    def test_object_get_or_create_tagfield(self):
        "Check that objects.get_or_create on TagField creates and gets"
        # Clear out models
        self.test_model.objects.all().delete()
        self.test_model.tags.tag_model.objects.all().delete()
        self.assertTagModel(self.test_model.tags, {})

        # Create
        t1, created = self.test_model.objects.get_or_create(tags="red, blue")
        self.assertEqual(created, True)
        self.assertTagModel(self.test_model.tags, {"red": 1, "blue": 1})
        t1.name = "Test 4"
        t1.save()
        self.assertEqual(str(t1.tags), "blue, red")

        # Get
        t2, created = self.test_model.objects.get_or_create(tags="red, blue")
        self.assertEqual(created, False)
        self.assertTagModel(self.test_model.tags, {"red": 1, "blue": 1})
        self.assertEqual(t2.name, "Test 4")
        self.assertEqual(str(t1.tags), "blue, red")

        self.assertEqual(t1.pk, t2.pk)

    #
    # .update_or_create()
    #

    def test_object_update_or_create_singletagfield(self):
        "Check that objects.update_or_create on SingleTagField creates and updates"
        # Clear out models
        self.test_model.objects.all().delete()
        self.test_model.singletag.tag_model.objects.all().delete()
        self.assertTagModel(self.test_model.singletag, {})

        # Create
        t1, created = self.test_model.objects.update_or_create(
            singletag="Mr", defaults={"name": "Created"}
        )
        self.assertEqual(created, True)
        self.assertTagModel(self.test_model.singletag, {"Mr": 1})

        # Get
        t2, created = self.test_model.objects.update_or_create(
            singletag="Mr", defaults={"name": "Updated"}
        )
        self.assertEqual(created, False)
        self.assertTagModel(self.test_model.singletag, {"Mr": 1})
        self.assertEqual(t2.name, "Updated")
        self.assertEqual(str(t2.singletag), "Mr")

        self.assertEqual(t1.pk, t2.pk)

    def test_object_update_or_create_tagfield(self):
        "Check that objects.update_or_create on TagField creates and gets"
        # Clear out models
        self.test_model.objects.all().delete()
        self.test_model.tags.tag_model.objects.all().delete()
        self.assertTagModel(self.test_model.tags, {})

        # Create
        t1, created = self.test_model.objects.update_or_create(
            tags="red, blue", defaults={"name": "Created"}
        )
        self.assertEqual(created, True)
        self.assertTagModel(self.test_model.tags, {"red": 1, "blue": 1})
        self.assertEqual(str(t1.tags), "blue, red")

        # Get
        t2, created = self.test_model.objects.update_or_create(
            tags="red, blue", defaults={"name": "Updated"}
        )
        self.assertEqual(created, False)
        self.assertTagModel(self.test_model.tags, {"red": 1, "blue": 1})
        self.assertEqual(t2.name, "Updated")
        self.assertEqual(str(t1.tags), "blue, red")

        self.assertEqual(t1.pk, t2.pk)

    #
    # .filter()
    #

    def test_object_singletag_filter(self):
        "Check that object.filter finds the correct items by singletag"
        qs1 = self.test_model.objects.filter(singletag="Mr")
        self.assertEqual(qs1.count(), 2)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o3.pk)

    def test_object_singletag_filter_misses(self):
        "Check that object.filter copes with missing singletag"
        qs1 = self.test_model.objects.filter(singletag="Miss")
        self.assertEqual(qs1.count(), 0)

    def test_object_tags_filter(self):
        "Check that object.filter finds the correct items by tags"
        qs1 = self.test_model.objects.filter(tags="red, blue")
        self.assertEqual(qs1.count(), 2)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o2.pk)

    def test_object_tags_filter_exact(self):
        "Check that object.filter finds the correct items by tags"
        qs1 = self.test_model.objects.filter(tags__exact="red, green, blue")
        self.assertEqual(qs1.count(), 2)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o2.pk)

    def test_object_tags_filter_exact_misses(self):
        "Check that object.filter only returns exact matches"
        qs1 = self.test_model.objects.filter(tags__exact="red, blue")
        self.assertEqual(qs1.count(), 0)

    def test_object_tags_filter_none(self):
        "Check that object.filter returns objects with no tags"
        o4 = self.test_model.objects.create(name="Test 4", tags="")
        qs1 = self.test_model.objects.filter(tags=None)
        self.assertEqual(qs1.count(), 1)
        self.assertEqual(qs1[0].pk, o4.pk)

    #
    # .exclude()
    #

    def test_object_singletag_exclude(self):
        "Check that object.exclude finds the correct items"
        qs1 = self.test_model.objects.exclude(singletag="Mrs")
        self.assertEqual(qs1.count(), 2)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o3.pk)

    def test_object_singletag_exclude_misses(self):
        "Check that object.exclude copes with missing singletag"
        qs1 = self.test_model.objects.exclude(singletag="Miss")
        self.assertEqual(qs1.count(), 3)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o2.pk)
        self.assertEqual(qs1[2].pk, self.o3.pk)

    def test_object_tags_exclude(self):
        "Check that object.exclude excludes when both are matched"
        qs1 = self.test_model.objects.exclude(tags="red, blue")
        self.assertEqual(qs1.count(), 1)
        self.assertEqual(qs1[0].pk, self.o3.pk)

    def test_object_tags_exclude_misses(self):
        "Check that object.exclude copes with missing tag"
        qs1 = self.test_model.objects.exclude(tags="red, yellow")
        self.assertEqual(qs1.count(), 3)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o2.pk)
        self.assertEqual(qs1[2].pk, self.o3.pk)

    def test_object_tags_exclude_exact(self):
        "Check that object.exclude exact excludes by tags"
        qs1 = self.test_model.objects.exclude(tags="red, green, blue")
        self.assertEqual(qs1.count(), 1)
        self.assertEqual(qs1[0].pk, self.o3.pk)

    def test_object_tags_exclude_exact_misses(self):
        "Check that object.exclude only excludes exact matches"
        qs1 = self.test_model.objects.exclude(tags="red, green, yellow")
        self.assertEqual(qs1.count(), 3)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o2.pk)
        self.assertEqual(qs1[2].pk, self.o3.pk)

    def test_object_tags_exclude_none(self):
        "Check that object.exclude returns objects with tags"
        self.test_model.objects.create(name="Test 4", tags="")
        qs1 = self.test_model.objects.exclude(tags=None)
        self.assertEqual(qs1.count(), 3)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o2.pk)
        self.assertEqual(qs1[2].pk, self.o3.pk)


@skip_if_mysql
class ModelTaggedQuerysetOptionsSingleTest(TagTestManager, TestCase):
    """
    Test tag options on tagged model querysets, for SingleTagField

    The only tag options which should be applied to queryset arguments is
    case_sensitive
    """

    manage_models = [test_models.SingleTagFieldOptionsModel]

    def setUpExtra(self):
        self.test_model = test_models.SingleTagFieldOptionsModel
        self.test_model.objects.create(
            name="Test 1", case_sensitive_true="Mr", case_sensitive_false="Mr"
        )
        self.test_model.objects.create(
            name="Test 2", case_sensitive_true="mr", case_sensitive_false="mr"
        )
        self.test_model.objects.create(
            name="Test 3", case_sensitive_true="Mr", case_sensitive_false="Mr"
        )

    def test_setup(self):
        "Confirm setup created objects as expected"
        self.assertTagModel(self.test_model.case_sensitive_true, {"Mr": 2, "mr": 1})
        self.assertTagModel(self.test_model.case_sensitive_false, {"Mr": 3})

    def test_case_sensitive_filter(self):
        "Check case sensitive matches"
        qs1 = self.test_model.objects.filter(case_sensitive_true="Mr")
        self.assertEqual(qs1.count(), 2)
        self.assertEqual(str(qs1[0].name), "Test 1")
        self.assertEqual(str(qs1[1].name), "Test 3")

    def test_case_sensitive_exclude_matches(self):
        "Check case sensitive excludes"
        qs1 = self.test_model.objects.exclude(case_sensitive_true="Mr")
        self.assertEqual(qs1.count(), 1)
        self.assertEqual(str(qs1[0].name), "Test 2")

    def test_case_insensitive_filter(self):
        "Check case insensitive matches"
        qs1 = self.test_model.objects.filter(case_sensitive_false="mr")
        self.assertEqual(qs1.count(), 3)
        self.assertEqual(str(qs1[0].name), "Test 1")
        self.assertEqual(str(qs1[1].name), "Test 2")
        self.assertEqual(str(qs1[2].name), "Test 3")

    def test_case_insensitive_exclude(self):
        "Check case insensitive matches"
        qs1 = self.test_model.objects.exclude(case_sensitive_false="mr")
        self.assertEqual(qs1.count(), 0)


@skip_if_mysql
class ModelTaggedQuerysetOptionsTest(TagTestManager, TestCase):
    """
    Test tag options on tagged model querysets, for TagField

    The only tag options which should be applied to queryset arguments is
    case_sensitive
    """

    manage_models = [test_models.TagFieldOptionsModel]

    def setUpExtra(self):
        self.test_model = test_models.TagFieldOptionsModel
        self.test_model.objects.create(
            name="Test 1", case_sensitive_true="Adam", case_sensitive_false="Adam"
        )
        self.test_model.objects.create(
            name="Test 2", case_sensitive_true="adam", case_sensitive_false="adam"
        )
        self.test_model.objects.create(
            name="Test 3", case_sensitive_true="Adam", case_sensitive_false="Adam"
        )

    def test_setup(self):
        "Confirm setup created objects as expected"
        self.assertTagModel(self.test_model.case_sensitive_true, {"Adam": 2, "adam": 1})
        self.assertTagModel(self.test_model.case_sensitive_false, {"Adam": 3})

    def test_case_sensitive_filter(self):
        "Check case sensitive matches"
        qs1 = self.test_model.objects.filter(case_sensitive_true="Adam")
        self.assertEqual(qs1.count(), 2)
        self.assertEqual(str(qs1[0].name), "Test 1")
        self.assertEqual(str(qs1[1].name), "Test 3")

    def test_case_sensitive_exclude_matches(self):
        "Check case sensitive excludes"
        qs1 = self.test_model.objects.exclude(case_sensitive_true="Adam")
        self.assertEqual(qs1.count(), 1)
        self.assertEqual(str(qs1[0].name), "Test 2")

    def test_case_insensitive_filter(self):
        "Check case insensitive matches"
        qs1 = self.test_model.objects.filter(case_sensitive_false="adam")
        self.assertEqual(qs1.count(), 3)
        self.assertEqual(str(qs1[0].name), "Test 1")
        self.assertEqual(str(qs1[1].name), "Test 2")
        self.assertEqual(str(qs1[2].name), "Test 3")

    def test_case_insensitive_exclude(self):
        "Check case insensitive matches"
        qs1 = self.test_model.objects.exclude(case_sensitive_false="adam")
        self.assertEqual(qs1.count(), 0)



"""
Tagulous test: Tag options

Modules tested:
    tagulous.options
    tagulous.constants
"""
from django.test import TestCase

from tagulous import constants as tag_constants
from tagulous import models as tag_models


class TagOptionsTest(TestCase):
    """
    Test TagOptions
    """

    def test_defaults(self):
        opt = tag_models.TagOptions()
        self.assertEqual(opt.items(with_defaults=False), {})
        self.assertEqual(opt.items(), tag_constants.OPTION_DEFAULTS)
        self.assertEqual(opt.form_items(with_defaults=False), {})
        self.assertEqual(
            opt.form_items(),
            dict(
                [
                    (k, v)
                    for k, v in tag_constants.OPTION_DEFAULTS.items()
                    if k in tag_constants.FORM_OPTIONS
                ]
            ),
        )

    def test_override_defaults(self):
        opt = tag_models.TagOptions(force_lowercase=True, case_sensitive=True)
        local_args = {"force_lowercase": True, "case_sensitive": True}
        self.assertEqual(opt.items(with_defaults=False), local_args)
        expected = {}
        expected.update(tag_constants.OPTION_DEFAULTS)
        expected.update(local_args)
        self.assertEqual(opt.items(), expected)
        self.assertEqual(opt.form_items(with_defaults=False), local_args)
        self.assertEqual(
            opt.form_items(),
            dict(
                [(k, v) for k, v in expected.items() if k in tag_constants.FORM_OPTIONS]
            ),
        )

    def test_initial_none(self):
        opt = tag_models.TagOptions(initial=None)
        self.assertEqual(opt.initial_string, "")
        self.assertEqual(opt.initial, [])

    def test_initial_string(self):
        opt = tag_models.TagOptions(initial="one, two")
        self.assertEqual(opt.initial_string, "one, two")
        self.assertEqual(opt.initial, ["one", "two"])

    def test_initial_list(self):
        opt = tag_models.TagOptions(initial=["one", "two"])
        self.assertEqual(opt.initial_string, "one, two")
        self.assertEqual(opt.initial, ["one", "two"])
        self.assertEqual(opt.force_lowercase, False)

    def test_set_invalid(self):
        with self.assertRaises(AttributeError) as cm:
            tag_models.TagOptions(invalid=False)
        self.assertEqual(str(cm.exception), "invalid")

    def test_get_invalid(self):
        opt = tag_models.TagOptions()
        with self.assertRaises(AttributeError) as cm:
            opt.invalid
        self.assertEqual(str(cm.exception), "invalid")

    def test_update_dict(self):
        opt = tag_models.TagOptions(initial="Adam, Brian")
        self.assertEqual(opt.initial_string, "Adam, Brian")

        opt.update({"initial": "Brian, Chris"})
        self.assertEqual(opt.initial_string, "Brian, Chris")

    def test_update_object(self):
        opt1 = tag_models.TagOptions(initial="Adam, Brian")
        opt2 = tag_models.TagOptions(initial="Brian, Chris")
        self.assertEqual(opt1.initial_string, "Adam, Brian")
        self.assertEqual(opt2.initial_string, "Brian, Chris")

        opt1.update(opt2)
        self.assertEqual(opt1.initial_string, "Brian, Chris")

    def test_set_missing_dict(self):
        opt = tag_models.TagOptions(initial="Adam, Brian")
        self.assertEqual(opt.initial_string, "Adam, Brian")
        self.assertEqual(opt.force_lowercase, False)

        opt.set_missing({"initial": "Brian, Chris", "force_lowercase": True})
        self.assertEqual(opt.initial_string, "Adam, Brian")
        self.assertEqual(opt.force_lowercase, True)

    def test_set_missing_object(self):
        opt1 = tag_models.TagOptions(initial="Adam, Brian")
        opt2 = tag_models.TagOptions(initial="Brian, Chris", force_lowercase=True)
        self.assertEqual(opt1.initial_string, "Adam, Brian")
        self.assertEqual(opt1.force_lowercase, False)
        self.assertEqual(opt2.initial_string, "Brian, Chris")
        self.assertEqual(opt2.force_lowercase, True)

        opt1.set_missing(opt2)
        self.assertEqual(opt1.initial_string, "Adam, Brian")
        self.assertEqual(opt1.force_lowercase, True)

    def test_add(self):
        opt1 = tag_models.TagOptions(initial="Adam, Brian")
        opt2 = tag_models.TagOptions(force_lowercase=True)
        self.assertEqual(opt1.initial_string, "Adam, Brian")
        self.assertEqual(opt1.force_lowercase, False)
        self.assertEqual(opt2.initial_string, "")
        self.assertEqual(opt2.force_lowercase, True)

        opt3 = opt1 + opt2
        self.assertEqual(opt3.initial_string, "Adam, Brian")
        self.assertEqual(opt3.force_lowercase, True)
        self.assertNotEqual(id(opt1), id(opt2))
        self.assertNotEqual(id(opt2), id(opt3))
        self.assertNotEqual(id(opt3), id(opt1))



"""
Tagulous default settings

Override these by setting new values in your global settings file
"""
from django.conf import settings


#
# Database control settings
#

NAME_MAX_LENGTH = getattr(settings, "TAGULOUS_NAME_MAX_LENGTH", 255)
SLUG_MAX_LENGTH = getattr(settings, "TAGULOUS_SLUG_MAX_LENGTH", 50)
LABEL_MAX_LENGTH = getattr(settings, "TAGULOUS_LABEL_MAX_LENGTH", NAME_MAX_LENGTH)

# Number of characters to allow for finding a unique slug, ie if set to 5, the
# slug will be truncated by up to 5 characters to allow for a suffix of _9999
SLUG_TRUNCATE_UNIQUE = getattr(settings, "TAGULOUS_SLUG_TRUNCATE_UNIQUE", 5)

# Set to false to generate ASCII slugs
SLUG_ALLOW_UNICODE = getattr(settings, "TAGULOUS_SLUG_ALLOW_UNICODE", False)


#
# Autocomplete settings
#

DEFAULT_AUTOCOMPLETE_JS = (
    "tagulous/lib/jquery.js",
    "tagulous/lib/select2-4/js/select2.full.min.js",
    "tagulous/tagulous.js",
    "tagulous/adaptor/select2-4.js",
)
DEFAULT_AUTOCOMPLETE_CSS = {"all": ["tagulous/lib/select2-4/css/select2.min.css"]}

AUTOCOMPLETE_JS = getattr(settings, "TAGULOUS_AUTOCOMPLETE_JS", DEFAULT_AUTOCOMPLETE_JS)
AUTOCOMPLETE_CSS = getattr(
    settings, "TAGULOUS_AUTOCOMPLETE_CSS", DEFAULT_AUTOCOMPLETE_CSS
)
AUTOCOMPLETE_SETTINGS = getattr(settings, "TAGULOUS_AUTOCOMPLETE_SETTINGS", None)

# Use vendored jquery and select2 for admin
DEFAULT_ADMIN_AUTOCOMPLETE_JS = (
    "tagulous/tagulous.js",
    "tagulous/adaptor/select2-4.js",
)
DEFAULT_ADMIN_AUTOCOMPLETE_CSS = {}

ADMIN_AUTOCOMPLETE_JS = getattr(
    settings, "TAGULOUS_ADMIN_AUTOCOMPLETE_JS", DEFAULT_ADMIN_AUTOCOMPLETE_JS
)
ADMIN_AUTOCOMPLETE_CSS = getattr(
    settings, "TAGULOUS_ADMIN_AUTOCOMPLETE_CSS", DEFAULT_ADMIN_AUTOCOMPLETE_CSS
)
ADMIN_AUTOCOMPLETE_SETTINGS = getattr(
    settings, "TAGULOUS_ADMIN_AUTOCOMPLETE_SETTINGS", AUTOCOMPLETE_SETTINGS
)


#
# Tag weighting defaults, for tag model queryset .weight() method
#

WEIGHT_MIN = getattr(settings, "TAGULOUS_WEIGHT_MIN", 1)
WEIGHT_MAX = getattr(settings, "TAGULOUS_WEIGHT_MAX", 6)


#
# Feature flags
#

# Option to automatically enhance Model, Manager and QuerySet classes so they
# know how to work with SingleTagFields and TagFields.
#
# This will automatically change models which use tag fields to subclass
# TaggedModel, their managers to subclass TaggedManager, and their querysets to
# subclass TaggedQuerySet
#
# If this is set to False, certain aspects of tag fields will not work as
# expected; you should consider manually subclassing the relevant classes.
# See settings documentation for more information.
ENHANCE_MODELS = getattr(settings, "TAGULOUS_ENHANCE_MODELS", True)



"""
Tagulous test: Tag Trees

Modules tested:
    tagulous.models.tree
"""
from django.db import IntegrityError
from django.test import TestCase, TransactionTestCase

import tagulous.settings as tagulous_settings
from tagulous import models as tag_models
from tests.lib import TagTestManager
from tests.tagulous_tests_app import models as test_models


class TagTreeTestManager(TagTestManager):
    """
    Manage tag tree tests
    """

    def assertTreeTag(
        self,
        tag,
        name=None,
        label=None,
        slug=None,
        path=None,
        parent=None,
        count=None,
        protected=None,
        level=None,
    ):
        "Check tag attributes match those specified"
        if name is not None:
            self.assertEqual(tag.name, name)
        if label is not None:
            self.assertEqual(tag.label, label)
        if slug is not None:
            self.assertEqual(tag.slug, slug)
        if path is not None:
            self.assertEqual(tag.path, path)

        if parent is not None:
            self.assertEqual(tag.parent, parent)
        if count is not None:
            self.assertEqual(tag.count, count)
        if protected is not None:
            self.assertEqual(tag.protected, protected)
        if level is not None:
            self.assertEqual(tag.level, level)


# ##############################################################################
# ###### TagTreeModel basics
# ##############################################################################


class TagTreeModelTest(TagTreeTestManager, TestCase):
    """
    Test TagTreeModel basics - that the models are created correctly, and tag
    instances can be created directly with parents and fields correct.

    These tests refer to levels; the higher the level, the deeper in the tree,
    eg: level1/level2/level3
    """

    manage_models = [test_models.TreeTest]

    def setUpExtra(self):
        self.test_model = test_models.TreeTest
        self.singletag_field = test_models.TreeTest.singletag
        self.singletag_model = test_models.TreeTest.singletag.tag_model
        self.tag_field = test_models.TreeTest.tags
        self.tag_model = test_models.TreeTest.tags.tag_model

    def test_label_field_length(self):
        """
        Value is initialized in setup.py runtests() settings
        """
        self.assertEqual(
            self.tag_model._meta.get_field("label").max_length,
            tagulous_settings.LABEL_MAX_LENGTH,
        )

    def test_singletag_model(self):
        "Check tag tree model is used as a base when SingleTagField tree=True"
        self.assertTrue(issubclass(self.singletag_model, tag_models.TagTreeModel))

    def test_tag_model(self):
        "Check tag tree model is used as a base when TagField tree=True"
        self.assertTrue(issubclass(self.tag_model, tag_models.TagTreeModel))

    def test_empty_name_raises_integrity_error(self):
        with self.assertRaises(IntegrityError) as cm:
            self.tag_model.objects.create(name=None)
        self.assertEqual(
            str(cm.exception),
            "NOT NULL constraint failed: tagulous_tests_app_tagulous_treetest_tags.name",
        )

    def test_level_1(self):
        "Check level 1 node created correctly"
        t1 = self.tag_model.objects.create(name="One")
        self.assertTreeTag(t1, name="One", label="One", slug="one", path="one", level=1)

    def test_level_1_single_character(self):
        "Check level 1 node created correctly if only one character"
        t1 = self.tag_model.objects.create(name="O")
        self.assertTreeTag(
            t1, name="O", label="O", slug="o", path="o", level=1,
        )

    def test_level_2_existing_l1(self):
        "Check level 2 node created with existing level 1"
        t1 = self.tag_model.objects.create(name="One")
        t2 = self.tag_model.objects.create(name="One/Two")
        self.assertTreeTag(
            t2,
            name="One/Two",
            label="Two",
            slug="two",
            path="one/two",
            parent=t1,
            level=2,
        )

    def test_level_2_missing_l1(self):
        "Check level 2 node creates missing level 1"
        t2 = self.tag_model.objects.create(name="One/Two")
        t1 = self.tag_model.objects.get(name="One")
        self.assertTreeTag(t1, name="One", label="One", slug="one", path="one", level=1)
        self.assertTreeTag(
            t2,
            name="One/Two",
            label="Two",
            slug="two",
            path="one/two",
            parent=t1,
            level=2,
        )

    def test_level_3_existing_l1_l2(self):
        "Check level 3 node created with existing level 1 and 2"
        self.tag_model.objects.create(name="One")
        t2 = self.tag_model.objects.create(name="One/Two")
        t3 = self.tag_model.objects.create(name="One/Two/Three")
        self.assertTreeTag(
            t3,
            name="One/Two/Three",
            label="Three",
            slug="three",
            path="one/two/three",
            parent=t2,
            level=3,
        )

    def test_level_3_existing_l1_missing_l2(self):
        "Check level 3 node created with existing level 1 but missing level 2"
        t1 = self.tag_model.objects.create(name="One")
        t3 = self.tag_model.objects.create(name="One/Two/Three")
        t2 = self.tag_model.objects.get(name="One/Two")
        self.assertTreeTag(
            t2,
            name="One/Two",
            label="Two",
            slug="two",
            path="one/two",
            parent=t1,
            level=2,
        )
        self.assertTreeTag(
            t3,
            name="One/Two/Three",
            label="Three",
            slug="three",
            path="one/two/three",
            parent=t2,
            level=3,
        )

    def test_level_3_missing_l1_l2(self):
        "Check level 3 node created with missing level 1 and 2"
        t3 = self.tag_model.objects.create(name="One/Two/Three")
        t1 = self.tag_model.objects.get(name="One")
        t2 = self.tag_model.objects.get(name="One/Two")
        self.assertTreeTag(t1, name="One", label="One", slug="one", path="one", level=1)
        self.assertTreeTag(
            t2,
            name="One/Two",
            label="Two",
            slug="two",
            path="one/two",
            parent=t1,
            level=2,
        )
        self.assertTreeTag(
            t3,
            name="One/Two/Three",
            label="Three",
            slug="three",
            path="one/two/three",
            parent=t2,
            level=3,
        )

    def test_level_1_renames_l2_l3(self):
        "Check renaming level 1 node renames l2 and l3"
        t3 = self.tag_model.objects.create(name="One/Two/Three")
        t1 = self.tag_model.objects.get(name="One")
        self.assertTagModel(
            self.tag_model, {"One": 0, "One/Two": 0, "One/Two/Three": 0}
        )
        t1.name = "Uno"
        t1.slug = None  # Force slug to be recreated
        t1.save()

        # Reload
        self.assertTagModel(
            self.tag_model, {"Uno": 0, "Uno/Two": 0, "Uno/Two/Three": 0}
        )
        t1 = self.tag_model.objects.get(name="Uno")
        t2 = self.tag_model.objects.get(name="Uno/Two")
        t3 = self.tag_model.objects.get(name="Uno/Two/Three")

        self.assertTreeTag(t1, name="Uno", label="Uno", slug="uno", path="uno", level=1)
        self.assertTreeTag(
            t2,
            name="Uno/Two",
            label="Two",
            slug="two",
            path="uno/two",
            parent=t1,
            level=2,
        )
        self.assertTreeTag(
            t3,
            name="Uno/Two/Three",
            label="Three",
            slug="three",
            path="uno/two/three",
            parent=t2,
            level=3,
        )

    def test_rebuild(self):
        "Check rebuild updates all slugs"
        # Break slugs
        t3 = self.tag_model.objects.create(name="One/Two/Three")
        for tag in self.tag_model.objects.all():
            tag.slug = str(tag.pk)
            tag.save()

        # Load
        t1 = self.tag_model.objects.get(name="One")
        t2 = self.tag_model.objects.get(name="One/Two")
        t3 = self.tag_model.objects.get(name="One/Two/Three")
        t1_slug = str(t1.pk)
        t2_slug = str(t2.pk)
        t3_slug = str(t3.pk)
        self.assertRegex(t1_slug, r"^\d+$")
        self.assertRegex(t2_slug, r"^\d+$")
        self.assertRegex(t3_slug, r"^\d+$")
        self.assertTreeTag(t1, name="One", slug=t1_slug, path=t1_slug)
        self.assertTreeTag(
            t2,
            name="One/Two",
            slug=t2_slug,
            path="{}/{}".format(t1_slug, t2_slug),
            parent=t1,
        )
        self.assertTreeTag(
            t3,
            name="One/Two/Three",
            slug=t3_slug,
            path="{}/{}/{}".format(t1_slug, t2_slug, t3_slug),
            parent=t2,
        )

        # Rebuild and re-test
        self.tag_model.objects.rebuild()
        t1 = self.tag_model.objects.get(name="One")
        t2 = self.tag_model.objects.get(name="One/Two")
        t3 = self.tag_model.objects.get(name="One/Two/Three")
        self.assertTreeTag(t1, name="One", slug="one", path="one")
        self.assertTreeTag(t2, name="One/Two", slug="two", path="one/two", parent=t1)
        self.assertTreeTag(
            t3, name="One/Two/Three", slug="three", path="one/two/three", parent=t2
        )


# ##############################################################################
# ###### TagTreeModel tree merging
# ##############################################################################


class TagTreeModelMergeTest(TagTreeTestManager, TestCase):
    """
    Test merging TagTreeModel nodes
    """

    manage_models = [test_models.TreeTest]

    def setUpExtra(self):
        self.test_model = test_models.TreeTest
        self.tag_field = test_models.TreeTest.tags
        self.tag_model = test_models.TreeTest.tags.tag_model

        # Normal Animal leaf nodes
        self.test_model.objects.create(name="cat1", tags="Animal/Mammal/Cat")
        self.test_model.objects.create(name="dog1", tags="Animal/Mammal/Dog")
        self.test_model.objects.create(name="bee1", tags="Animal/Insect/Bee")

        # Pet tree has Cat leaf and extra L2 not in Animal
        self.test_model.objects.create(name="cat2", tags="Pet/Mammal/Cat")
        self.test_model.objects.create(name="robin1", tags="Pet/Bird/Robin")

        # Food tree has Cat leaf and extra L3 not in Animal
        self.test_model.objects.create(name="cat3", tags="Food/Mammal/Cat")
        self.test_model.objects.create(name="pig1", tags="Food/Mammal/Pig")

    def test_setup(self):
        "Check test setup is correct"
        # There should be 6 tags
        self.assertTagModel(
            self.tag_model,
            {
                "Animal": 0,
                "Animal/Insect": 0,
                "Animal/Insect/Bee": 1,
                "Animal/Mammal": 0,
                "Animal/Mammal/Cat": 1,
                "Animal/Mammal/Dog": 1,
                "Pet": 0,
                "Pet/Mammal": 0,
                "Pet/Mammal/Cat": 1,
                "Pet/Bird": 0,
                "Pet/Bird/Robin": 1,
                "Food": 0,
                "Food/Mammal": 0,
                "Food/Mammal/Cat": 1,
                "Food/Mammal/Pig": 1,
            },
        )

    def test_merge_l1(self):
        """
        Test that merging an l1 without merging children reassigns tagged
        items, but leaves self and children
        """
        # Tag something with 'Pet'
        t1 = self.test_model.objects.create(name="pet1", tags="Pet")
        self.assertEqual(self.tag_model.objects.get(name="Pet").count, 1)

        # Now merge
        self.tag_model.objects.get(name="Animal").merge_tags(
            "Pet, Food", children=False
        )
        self.assertTagModel(
            self.tag_model,
            {
                "Animal": 1,
                "Animal/Insect": 0,
                "Animal/Insect/Bee": 1,
                "Animal/Mammal": 0,
                "Animal/Mammal/Cat": 1,
                "Animal/Mammal/Dog": 1,
                "Pet": 0,
                "Pet/Mammal": 0,
                "Pet/Mammal/Cat": 1,
                "Pet/Bird": 0,
                "Pet/Bird/Robin": 1,
                "Food": 0,
                "Food/Mammal": 0,
                "Food/Mammal/Cat": 1,
                "Food/Mammal/Pig": 1,
            },
        )
        self.assertInstanceEqual(t1, name="pet1", tags="Animal")

    def test_merge_l1_with_children(self):
        """
        Test that merging an l1 and its children merges the full subtrees and
        cleans empty parents
        """
        # Now merge
        self.tag_model.objects.get(name="Animal").merge_tags("Pet, Food", children=True)
        self.assertTagModel(
            self.tag_model,
            {
                "Animal": 0,
                "Animal/Insect": 0,
                "Animal/Insect/Bee": 1,
                "Animal/Mammal": 0,
                "Animal/Mammal/Cat": 3,
                "Animal/Mammal/Dog": 1,
                "Animal/Mammal/Pig": 1,
                "Animal/Bird": 0,
                "Animal/Bird/Robin": 1,
            },
        )

    def test_merge_l2(self):
        """
        Test that merging an l2 without merging children reassigns tagged
        items, but leaves self and children
        """
        # Tag something with 'Pet/Mammal'
        t1 = self.test_model.objects.create(name="mammal1", tags="Pet/Mammal")
        self.assertEqual(self.tag_model.objects.get(name="Pet/Mammal").count, 1)

        # Now merge
        self.tag_model.objects.get(name="Animal/Mammal").merge_tags(
            "Pet/Mammal, Food/Mammal", children=False
        )
        self.assertTagModel(
            self.tag_model,
            {
                "Animal": 0,
                "Animal/Insect": 0,
                "Animal/Insect/Bee": 1,
                "Animal/Mammal": 1,
                "Animal/Mammal/Cat": 1,
                "Animal/Mammal/Dog": 1,
                "Pet": 0,
                "Pet/Mammal": 0,
                "Pet/Mammal/Cat": 1,
                "Pet/Bird": 0,
                "Pet/Bird/Robin": 1,
                "Food": 0,
                "Food/Mammal": 0,
                "Food/Mammal/Cat": 1,
                "Food/Mammal/Pig": 1,
            },
        )
        self.assertInstanceEqual(t1, name="mammal1", tags="Animal/Mammal")

    def test_merge_l2_with_children(self):
        """
        Test that merging an l2 and its children merges the full subtrees and
        cleans empty parents
        """
        # Now merge
        self.tag_model.objects.get(name="Animal/Mammal").merge_tags(
            "Pet/Mammal, Food/Mammal", children=True
        )
        self.assertTagModel(
            self.tag_model,
            {
                "Animal": 0,
                "Animal/Insect": 0,
                "Animal/Insect/Bee": 1,
                "Animal/Mammal": 0,
                "Animal/Mammal/Cat": 3,
                "Animal/Mammal/Dog": 1,
                "Animal/Mammal/Pig": 1,
                "Pet": 0,
                "Pet/Bird": 0,
                "Pet/Bird/Robin": 1,
            },
        )

    def test_merge_l3(self):
        "Test merging an l3 cleans empty parents"
        # No need to test children=True - there are no children
        self.tag_model.objects.get(name="Animal/Mammal/Cat").merge_tags(
            "Pet/Mammal/Cat, Food/Mammal/Cat", children=False
        )
        self.assertTagModel(
            self.tag_model,
            {
                "Animal": 0,
                "Animal/Insect": 0,
                "Animal/Insect/Bee": 1,
                "Animal/Mammal": 0,
                "Animal/Mammal/Cat": 3,
                "Animal/Mammal/Dog": 1,
                "Pet": 0,
                # Pet/Mammal now empty
                "Pet/Bird": 0,
                "Pet/Bird/Robin": 1,
                "Food": 0,
                "Food/Mammal": 0,
                "Food/Mammal/Pig": 1,
            },
        )


# ##############################################################################
# ###### Custom TagTreeModel basics
# ##############################################################################


class TagTreeModelCustomTest(TagTreeTestManager, TransactionTestCase):
    """
    Test TagTreeModel basics when using a custom model

    Use a TransactionTestCase so the apps will be reset
    """

    manage_models = [test_models.CustomTreeTest]

    def setUpExtra(self):
        self.test_model = test_models.CustomTreeTest
        self.tag_model = test_models.CustomTagTree
        self.singletag_field = test_models.CustomTreeTest.singletag
        self.tag_field = test_models.CustomTreeTest.tags

    def test_tag_model(self):
        "Check tag tree model is used as a base"
        self.assertTrue(issubclass(self.tag_model, tag_models.TagTreeModel))

    def test_tree_option(self):
        "Check tag tree model forces tree=True when missing"
        self.assertTrue(self.tag_model.tag_options.tree)
        self.assertTrue(self.singletag_field.tag_options.tree)
        self.assertTrue(self.tag_field.tag_options.tree)

    def test_tree_option_tagmodel_fail(self):
        "Check tree option is not allowed in TagMeta"
        with self.assertRaises(ValueError) as cm:

            class FailModel_tree_tagmeta(tag_models.TagModel):
                class TagMeta:
                    tree = True

        self.assertEqual(str(cm.exception), "Cannot set tree option in TagMeta")


# ##############################################################################
# ###### TagTreeModel tree navigation methods
# ##############################################################################


class TagTreeModelNavTest(TagTreeTestManager, TestCase):
    """
    Test navigation through TagTreeModel - get_ancestors etc
    """

    manage_models = [test_models.TreeTest]

    def setUpExtra(self):
        self.tag_field = test_models.TreeTest.tags
        self.tag_model = test_models.TreeTest.tags.tag_model

        self.cat = self.tag_model.objects.create(name="Animal/Mammal/Cat")
        self.dog = self.tag_model.objects.create(name="Animal/Mammal/Dog")
        self.bee = self.tag_model.objects.create(name="Animal/Insect/Bee")

    def test_setup(self):
        "Check test setup is correct"
        # There should be 6 tags
        self.assertTagModel(
            self.tag_model,
            {
                "Animal": 0,
                "Animal/Insect": 0,
                "Animal/Insect/Bee": 0,
                "Animal/Mammal": 0,
                "Animal/Mammal/Cat": 0,
                "Animal/Mammal/Dog": 0,
            },
        )

    def test_ancestors(self):
        "Check l1 and l2 ancestors found from l3"
        # Add in some more extras to make sure they're not picked up
        self.tag_model.objects.create(name="Fail")
        self.tag_model.objects.create(name="Animal/Fail")
        self.tag_model.objects.create(name="Animal/Mammal/Cat/Fail")

        # Look up from Cat
        t1 = self.tag_model.objects.get(name="Animal/Mammal/Cat")
        anc = t1.get_ancestors()
        self.assertEqual(len(anc), 2)
        self.assertTreeTag(anc[0], name="Animal", level=1)
        self.assertTreeTag(anc[1], name="Animal/Mammal", level=2)

    def test_no_ancestors(self):
        "Check no ancestors found from l1"
        t1 = self.tag_model.objects.get(name="Animal")
        self.assertEqual(len(t1.get_ancestors()), 0)

    def test_descendants_l1(self):
        "Check l2 and l3 descendants found from l1"
        # Add in some extras to make sure they're not picked up
        self.tag_model.objects.create(name="Fail")

        # Look down from Animal
        t1 = self.tag_model.objects.get(name="Animal")
        self.assertTreeTag(t1, name="Animal", level=1)
        dec = t1.get_descendants()
        self.assertEqual(len(dec), 5)
        self.assertTreeTag(dec[0], name="Animal/Insect", level=2)
        self.assertTreeTag(dec[1], name="Animal/Insect/Bee", level=3)
        self.assertTreeTag(dec[2], name="Animal/Mammal", level=2)
        self.assertTreeTag(dec[3], name="Animal/Mammal/Cat", level=3)
        self.assertTreeTag(dec[4], name="Animal/Mammal/Dog", level=3)

    def test_descendants_l3_from_l2(self):
        "Check l3 descendants found from l2"
        # Look down from Animal
        t1 = self.tag_model.objects.get(name="Animal/Mammal")
        self.assertTreeTag(t1, name="Animal/Mammal", level=2)
        dec = t1.get_descendants()
        self.assertEqual(len(dec), 2)
        self.assertTreeTag(dec[0], name="Animal/Mammal/Cat", level=3)
        self.assertTreeTag(dec[1], name="Animal/Mammal/Dog", level=3)

    def test_no_descendants_from_l3(self):
        "Check no descendants found from l3"
        # Look down from Animal
        t1 = self.tag_model.objects.get(name="Animal/Insect/Bee")
        self.assertTreeTag(t1, name="Animal/Insect/Bee", level=3)
        dec = t1.get_descendants()
        self.assertEqual(len(dec), 0)

    def test_siblings_l1(self):
        "Find level 1 siblings"
        # Add another level 1 tag to find
        self.tag_model.objects.create(name="Vegetable")

        t1 = self.tag_model.objects.get(name="Animal")
        sibs = t1.get_siblings()
        self.assertEqual(len(sibs), 2)
        self.assertTreeTag(sibs[0], name="Animal", level=1)
        self.assertTreeTag(sibs[1], name="Vegetable", level=1)

    def test_siblings_l2(self):
        "Find level 2 siblings"
        t1 = self.tag_model.objects.get(name="Animal/Insect")
        sibs = t1.get_siblings()
        self.assertEqual(len(sibs), 2)
        self.assertTreeTag(sibs[0], name="Animal/Insect", level=2)
        self.assertTreeTag(sibs[1], name="Animal/Mammal", level=2)


# ##############################################################################
# ###### TagTreeQuerySet
# ##############################################################################


class TagTreeQuerySetTest(TagTreeTestManager, TestCase):
    """
    Test navigation using TagTreeModelQuerySet
    """

    manage_models = [test_models.TreeTest]

    def setUpExtra(self):
        self.tag_field = test_models.TreeTest.tags
        self.tag_model = test_models.TreeTest.tags.tag_model

        self.cat = self.tag_model.objects.create(name="Animal/Mammal/Cat")
        self.dog = self.tag_model.objects.create(name="Animal/Mammal/Dog")
        self.bee = self.tag_model.objects.create(name="Animal/Insect/Bee")
        self.vegetable = self.tag_model.objects.create(name="Vegetable")

    def test_setup(self):
        "Check test setup is correct"
        # There should be 6 tags
        self.assertTagModel(
            self.tag_model,
            {
                "Animal": 0,
                "Animal/Insect": 0,
                "Animal/Insect/Bee": 0,
                "Animal/Mammal": 0,
                "Animal/Mammal/Cat": 0,
                "Animal/Mammal/Dog": 0,
                "Vegetable": 0,
            },
        )

    def test_ancestors_l1(self):
        qs = self.tag_model.objects.filter(level=1)
        self.assertSequenceEqual(
            qs.values_list("label", flat=True), ["Animal", "Vegetable"]
        )

        qs = qs.with_ancestors()
        self.assertSequenceEqual(
            qs.values_list("name", flat=True), ["Animal", "Vegetable"]
        )

    def test_ancestors_l2(self):
        qs = self.tag_model.objects.filter(level=2)
        self.assertSequenceEqual(
            qs.values_list("label", flat=True), ["Insect", "Mammal"]
        )

        qs = qs.with_ancestors()
        self.assertSequenceEqual(
            qs.values_list("name", flat=True),
            ["Animal", "Animal/Insect", "Animal/Mammal"],
        )

    def test_ancestors_l3(self):
        qs = self.tag_model.objects.filter(level=3)
        self.assertSequenceEqual(
            qs.values_list("label", flat=True), ["Bee", "Cat", "Dog"]
        )

        qs = qs.with_ancestors()
        self.assertSequenceEqual(
            qs.values_list("name", flat=True),
            [
                "Animal",
                "Animal/Insect",
                "Animal/Insect/Bee",
                "Animal/Mammal",
                "Animal/Mammal/Cat",
                "Animal/Mammal/Dog",
            ],
        )

    def test_descendants_l1(self):
        qs = self.tag_model.objects.filter(level=1)
        self.assertSequenceEqual(
            qs.values_list("label", flat=True), ["Animal", "Vegetable"]
        )

        qs = qs.with_descendants()
        self.assertSequenceEqual(
            qs.values_list("name", flat=True),
            [
                "Animal",
                "Animal/Insect",
                "Animal/Insect/Bee",
                "Animal/Mammal",
                "Animal/Mammal/Cat",
                "Animal/Mammal/Dog",
                "Vegetable",
            ],
        )

    def test_descendants_l2(self):
        qs = self.tag_model.objects.filter(level=2)
        self.assertSequenceEqual(
            qs.values_list("label", flat=True), ["Insect", "Mammal"]
        )

        qs = qs.with_descendants()
        self.assertSequenceEqual(
            qs.values_list("name", flat=True),
            [
                "Animal/Insect",
                "Animal/Insect/Bee",
                "Animal/Mammal",
                "Animal/Mammal/Cat",
                "Animal/Mammal/Dog",
            ],
        )

    def test_descendants_l3(self):
        qs = self.tag_model.objects.filter(level=3)
        self.assertSequenceEqual(
            qs.values_list("label", flat=True), ["Bee", "Cat", "Dog"]
        )

        qs = qs.with_descendants()
        self.assertSequenceEqual(
            qs.values_list("name", flat=True),
            ["Animal/Insect/Bee", "Animal/Mammal/Cat", "Animal/Mammal/Dog"],
        )

    def test_siblings_l1(self):
        qs = self.tag_model.objects.filter(name="Animal")
        self.assertSequenceEqual(qs.values_list("label", flat=True), ["Animal"])

        qs = qs.with_siblings()
        self.assertSequenceEqual(
            qs.values_list("name", flat=True), ["Animal", "Vegetable"]
        )

    def test_siblings_l2(self):
        # Throw in something else to ensure l2 stays within subtree
        self.tag_model.objects.create(name="Vegetable/Horrible")

        qs = self.tag_model.objects.filter(name="Animal/Insect")
        self.assertSequenceEqual(qs.values_list("label", flat=True), ["Insect"])

        qs = qs.with_siblings()
        self.assertSequenceEqual(
            qs.values_list("name", flat=True), ["Animal/Insect", "Animal/Mammal"]
        )

    def test_siblings_l3(self):
        # Throw in something else to ensure l3 stays within subtree
        self.tag_model.objects.create(name="Vegetable/Horrible/Mushroom")

        qs = self.tag_model.objects.filter(name="Animal/Mammal/Cat")
        self.assertSequenceEqual(qs.values_list("label", flat=True), ["Cat"])

        qs = qs.with_siblings()
        self.assertSequenceEqual(
            qs.values_list("name", flat=True),
            ["Animal/Mammal/Cat", "Animal/Mammal/Dog"],
        )


# ##############################################################################
# ###### TagTreeModel access via fields
# ##############################################################################


class TagTreeModelFieldTest(TagTreeTestManager, TestCase):
    """
    Test TagTreeModel access via fields - creation and deletion
    """

    manage_models = [test_models.TreeTest]

    def setUpExtra(self):
        self.singletag_field = test_models.TreeTest.singletag
        self.singletag_model = test_models.TreeTest.singletag.tag_model
        self.tag_field = test_models.TreeTest.tags
        self.tag_model = test_models.TreeTest.tags.tag_model

    def test_singletagfield_level_3(self):
        """
        Check level 3 nodes created with missing level 1 and 2 when assigned
        via singletag fields
        """
        test_models.TreeTest.objects.create(name="Test 1", singletag="One/Two/Three")
        t1 = self.singletag_model.objects.get(name="One")
        t2 = self.singletag_model.objects.get(name="One/Two")
        t3 = self.singletag_model.objects.get(name="One/Two/Three")
        self.assertTreeTag(t1, name="One", label="One", slug="one", path="one", level=1)
        self.assertTreeTag(
            t2,
            name="One/Two",
            label="Two",
            slug="two",
            path="one/two",
            parent=t1,
            level=2,
        )
        self.assertTreeTag(
            t3,
            name="One/Two/Three",
            label="Three",
            slug="three",
            path="one/two/three",
            parent=t2,
            count=1,
            level=3,
        )

        self.assertTagModel(
            self.singletag_model, {"One": 0, "One/Two": 0, "One/Two/Three": 1}
        )
        self.assertTagModel(self.tag_model, {})

    def test_tagfield_level_3(self):
        """
        Check level 3 nodes created with missing level 1 and 2 when assigned
        via tag fields
        """
        test_models.TreeTest.objects.create(
            name="Test 1", tags="One/Two/Three, Uno/Dos/Tres"
        )

        self.assertTagModel(self.singletag_model, {})
        self.assertTagModel(
            self.tag_model,
            {
                "One": 0,
                "One/Two": 0,
                "One/Two/Three": 1,
                "Uno": 0,
                "Uno/Dos": 0,
                "Uno/Dos/Tres": 1,
            },
        )

        t1 = self.tag_model.objects.get(name="One")
        t2 = self.tag_model.objects.get(name="One/Two")
        t3 = self.tag_model.objects.get(name="One/Two/Three")
        self.assertTreeTag(t1, name="One", label="One", slug="one", path="one", level=1)
        self.assertTreeTag(
            t2,
            name="One/Two",
            label="Two",
            slug="two",
            path="one/two",
            parent=t1,
            level=2,
        )
        self.assertTreeTag(
            t3,
            name="One/Two/Three",
            label="Three",
            slug="three",
            path="one/two/three",
            parent=t2,
            level=3,
            count=1,
        )

    def test_delete_l2_empty_l1(self):
        """
        Check deleting a level 2 node cleans away empty level 1
        """
        obj1 = test_models.TreeTest.objects.create(name="Test 1", singletag="One/Two")
        self.assertTagModel(self.singletag_model, {"One": 0, "One/Two": 1})

        obj1.singletag = ""
        obj1.save()
        self.assertTagModel(self.singletag_model, {})

    def test_delete_l2_populated_l1(self):
        """
        Check deleting a level 2 node leaves a populated l1
        """
        obj1 = test_models.TreeTest.objects.create(name="Test 1", singletag="Uno/Dos")
        self.singletag_model.objects.create(name="Uno/Due")
        self.assertTagModel(
            self.singletag_model, {"Uno": 0, "Uno/Dos": 1, "Uno/Due": 0}
        )

        obj1.singletag = ""
        obj1.save()
        self.assertTagModel(self.singletag_model, {"Uno": 0, "Uno/Due": 0})

    def test_delete_l3_populated_l2(self):
        """
        Check deleting a level 3 node leaves a populated level 2
        """
        obj1 = test_models.TreeTest.objects.create(
            name="Test 1", singletag="No/Uno/Dos"
        )
        self.singletag_model.objects.create(name="No/Uno/Due")
        self.assertTagModel(
            self.singletag_model,
            {"No": 0, "No/Uno": 0, "No/Uno/Dos": 1, "No/Uno/Due": 0},
        )

        obj1.singletag = ""
        obj1.save()
        self.assertTagModel(
            self.singletag_model, {"No": 0, "No/Uno": 0, "No/Uno/Due": 0}
        )

    def test_delete_l3_empty_l2_populated_l1(self):
        """
        Check deleting a level 3 node cleans empty l2 but leaves populated l1
        """
        obj1 = test_models.TreeTest.objects.create(
            name="Test 1", singletag="Uno/Dos/Tres"
        )
        self.singletag_model.objects.create(name="Uno/Due/Tre")
        self.assertTagModel(
            self.singletag_model,
            {"Uno": 0, "Uno/Dos": 0, "Uno/Dos/Tres": 1, "Uno/Due": 0, "Uno/Due/Tre": 0},
        )

        obj1.singletag = ""
        obj1.save()
        self.assertTagModel(
            self.singletag_model, {"Uno": 0, "Uno/Due": 0, "Uno/Due/Tre": 0}
        )

    def test_delete_l3_used_l2(self):
        """
        Check deleting a level 3 node leaves an empty but used level 2
        """
        test_models.TreeTest.objects.create(name="Test 1", singletag="Uno/Dos")
        obj2 = test_models.TreeTest.objects.create(
            name="Test 2", singletag="Uno/Dos/Tres"
        )
        self.assertTagModel(
            self.singletag_model, {"Uno": 0, "Uno/Dos": 1, "Uno/Dos/Tres": 1}
        )

        obj2.singletag = ""
        obj2.save()
        self.assertTagModel(self.singletag_model, {"Uno": 0, "Uno/Dos": 1})

    def test_delete_l1_used_l2(self):
        "Check deleting a level 1 node with used level 2 leaves both"
        obj1 = test_models.TreeTest.objects.create(name="Test 1", singletag="one")
        test_models.TreeTest.objects.create(name="Test 2", singletag="one/two")

        self.assertTagModel(self.singletag_model, {"one": 1, "one/two": 1})
        obj1.singletag = ""
        obj1.save()
        self.assertTagModel(self.singletag_model, {"one": 0, "one/two": 1})

    def test_get_descendant_count(self):
        "Check the count of the descendants"
        # Make four of everything
        for i in range(4):
            test_models.TreeTest.objects.create(name="Test 1.%d" % i, singletag="One")
            test_models.TreeTest.objects.create(
                name="Test 2.%d" % i, singletag="One/Two"
            )
            test_models.TreeTest.objects.create(
                name="Test 3.%d" % i, singletag="One/Two/Three"
            )
        self.assertTagModel(
            self.singletag_model, {"One": 4, "One/Two": 4, "One/Two/Three": 4}
        )

        # Check the counts
        t1 = self.singletag_model.objects.get(name="One")
        t2 = self.singletag_model.objects.get(name="One/Two")
        t3 = self.singletag_model.objects.get(name="One/Two/Three")
        self.assertEqual(t1.descendant_count, 8)
        self.assertEqual(t2.descendant_count, 4)
        self.assertEqual(t3.descendant_count, 0)

    def test_get_family_count(self):
        "Check the count of the family - node plus descendants"
        # Make four of everything
        for i in range(4):
            test_models.TreeTest.objects.create(name="Test 1.%d" % i, singletag="One")
            test_models.TreeTest.objects.create(
                name="Test 2.%d" % i, singletag="One/Two"
            )
            test_models.TreeTest.objects.create(
                name="Test 3.%d" % i, singletag="One/Two/Three"
            )
        self.assertTagModel(
            self.singletag_model, {"One": 4, "One/Two": 4, "One/Two/Three": 4}
        )

        # Check the counts
        t1 = self.singletag_model.objects.get(name="One")
        t2 = self.singletag_model.objects.get(name="One/Two")
        t3 = self.singletag_model.objects.get(name="One/Two/Three")
        self.assertEqual(t1.family_count, 12)
        self.assertEqual(t2.family_count, 8)
        self.assertEqual(t3.family_count, 4)



"""
Test models with capitalised names in the database (#60)
"""
from django.test import TestCase

from tests.lib import TagTestManager
from tests.tagulousTestsApp3 import models as test_models


class CapitalisedModelTest(TagTestManager, TestCase):
    """
    Test model with capitalised table name
    """

    manage_models = [test_models.CapitalisedTest]

    def setUpExtra(self):
        self.test_model = test_models.CapitalisedTest
        self.single_tag_model = test_models.CapitalisedTest.singletag.tag_model
        self.single_tag_field = test_models.CapitalisedTest.singletag
        self.tag_model = test_models.CapitalisedTest.tags.tag_model
        self.tag_field = test_models.CapitalisedTest.tags

    def test_table_name_capitalised(self):
        "Check table name is capitalised"
        # since that's what we're here to test
        self.assertEqual(
            self.test_model._meta.db_table, "tagulousTestsApp3_capitalisedtest"
        )
        self.assertEqual(
            self.single_tag_model._meta.db_table,
            "tagulousTestsApp3_tagulous_capitalisedtest_singletag",
        )
        self.assertEqual(
            self.tag_model._meta.db_table,
            "tagulousTestsApp3_tagulous_capitalisedtest_tags",
        )

    def test_tag_basics(self):
        "Check a tag string can be set and get"
        t1 = self.test_model.objects.create(
            name="Test 1", singletag="single1", tags="tag1, tag2"
        )
        self.assertInstanceEqual(
            t1, name="Test 1", singletag="single1", tags="tag1, tag2"
        )

    def test_weight(self):
        "Test weight()"
        # Copy of test_models_tagmodel.TagModelQuerySetTest.test_weight_scale_up
        self.o1 = self.test_model.objects.create(
            name="Test 1", singletag="single1", tags="David, Eric"
        )
        self.o2 = self.test_model.objects.create(
            name="Test 2", singletag="single1", tags="Eric, Frank"
        )

        # Mimic initial from TagFieldOptionsModel
        self.tag_model.objects.create(name="Adam")
        self.tag_model.objects.create(name="Brian")
        self.tag_model.objects.create(name="Chris")

        # Scale them to 2+2n: 0=2, 1=4, 2=6
        weighted = self.tag_model.objects.weight(min=2, max=6)
        self.assertEqual(len(weighted), 6)
        self.assertEqual(weighted[0].name, "Adam")
        self.assertEqual(weighted[0].weight, 2)

        self.assertEqual(weighted[1], "Brian")
        self.assertEqual(weighted[1].weight, 2)

        self.assertEqual(weighted[2], "Chris")
        self.assertEqual(weighted[2].weight, 2)

        self.assertEqual(weighted[3], "David")
        self.assertEqual(weighted[3].weight, 4)

        self.assertEqual(weighted[4], "Eric")
        self.assertEqual(weighted[4].weight, 6)

        self.assertEqual(weighted[5], "Frank")
        self.assertEqual(weighted[5].weight, 4)



"""
Tagulous model fields

These are the fields which users will add to models. During the
``contribute_to_class`` phase of model initialisation, they create any
necessary tag models, and add the descriptors back onto the model.

They are also responsible for preparing form fields.
"""
from django.core.checks import Warning as ChecksWarning
from django.db import models
from django.utils.text import capfirst

from .. import constants, forms
from .descriptors import SingleTagDescriptor, TagDescriptor
from .models import BaseTagModel, TagModel, TagTreeModel
from .options import TagOptions


# ##############################################################################
# ###### Mixin for model tag fields
# ##############################################################################


class BaseTagField(object):
    """
    Mixin for TagField and SingleTagField
    """

    # List of fields which are forbidden from __init__
    forbidden_fields = ()

    def __init__(self, to=None, to_base=None, **kwargs):
        """
        Initialise the tag options and store

        Takes all tag options as keyword arguments, as long as there is no
        ``to`` tag model specified.

        Undocumented keyword argument ``_set_tag_meta`` is used internally to
        update tag model options with those specified in keyword arguments when
        a ``to`` tag model is specified. However, this is intended for internal
        use only (when migrating), so if you must use it, use it with care.
        """
        # Save tag model data
        self.tag_model = to
        self.tag_model_base = to_base

        # Extract options from kwargs
        options = {}
        for key, default in constants.OPTION_DEFAULTS.items():
            # Look in kwargs, then in tag_meta
            if key in kwargs:
                options[key] = kwargs.pop(key)

        # Detect if _set_tag_meta is set
        set_tag_meta = kwargs.pop("_set_tag_meta", False)

        # Detect whether we need to automatically generate a tag model
        if self.tag_model:
            self.auto_tag_model = False

            # Tag model might be a string - set options later
            self.tag_options = None
            self._deferred_options = [set_tag_meta, options]

        else:
            self.auto_tag_model = True
            self.tag_options = TagOptions(**options)
            self._deferred_options = None

        # If the tag model was not specified, we need to specify one.
        # However, we can't reliably auto-generate a unique and repeatable
        # model name for tags here in __init__ - we can only do that in
        # contribute_to_class once we know the name of the field in the model.
        # We'll therefore use the string '-'; Django will not do anything about
        # resolving it until contribute_to_class, at which point we'll replace
        # it with a reference to the real tag model.
        kwargs["to"] = self.tag_model if self.tag_model else "-"

        # Call super __init__
        super(BaseTagField, self).__init__(**kwargs)

        # Make a note that this has not been contributed to a class yet
        self.contributed = False

    def do_related_class(self, other, cls):
        """
        Process tag model now it has been resolved if it was a string
        """
        # Set up relation as normal
        super(BaseTagField, self).do_related_class(other, cls)

        # Make sure tag model is the related model, in case it was a string
        self.tag_model = self.remote_field.model

        # Check class type of tag model
        if not issubclass(self.tag_model, BaseTagModel):
            raise ValueError("Tag model must be a subclass of TagModel")

        # Process the deferred options
        self._process_deferred_options(is_to_self=self.tag_model == cls)

    def _process_deferred_options(self, is_to_self=False):
        """
        Process tag options once we have the related model

        If the field is explicitly referring to itself, is_to_self will be True
        """
        # See if tag options were deferred
        if self._deferred_options is None:
            return

        # Get deferred options
        set_tag_meta, options = self._deferred_options
        self._deferred_options = None

        # Set options
        if options:
            if set_tag_meta:
                # Tag option arguments must be used to update the tag model
                # (and any other tag fields which use it) - used during
                # migration when the tag model doesn't know its own options
                #
                # However, to avoid issues with inheritance (particularly during
                # migrations), clone the tag_options rather than changing in place
                tag_options = self.tag_model.tag_options.clone()
                self.tag_model.tag_options = tag_options.update(options)
            else:
                raise ValueError(
                    "Cannot set tag options on explicit tag model %r"
                    % (self.tag_model,)
                )

        # Link to model options by reference in case they get updated later
        self.tag_options = self.tag_model.tag_options

    def contribute_to_class(self, cls, name):
        """
        Create the tag model if necessary, then initialise and contribute the
        field to the class
        """
        #
        # Get or create the tag model
        #

        # Create a new tag model if we need to
        if self.auto_tag_model:
            # Generate a list of attributes for the new tag model
            model_attrs = {
                # Module should be the same as the main model
                "__module__": cls.__module__,
                # Give it access to the options
                "tag_options": self.tag_options,
            }

            # Build new tag model
            # Name is Tagulous_MODELNAME_FIELDNAME
            model_name = "%s_%s_%s" % (
                constants.MODEL_PREFIX,
                cls._meta.object_name,
                name,
            )

            if self.tag_model_base is not None:
                model_cls = self.tag_model_base
            elif self.tag_options.tree:
                model_cls = TagTreeModel
            else:
                model_cls = TagModel

            self.tag_model = type(model_name, (model_cls,), model_attrs)

            # Give it a verbose name, for admin filters
            verbose_name_singular = (
                self.tag_options.verbose_name_singular or self.verbose_name or name
            )
            verbose_name_plural = self.tag_options.verbose_name_plural
            if not verbose_name_plural:
                verbose_name_plural = verbose_name_singular or self.verbose_name or name
                if not verbose_name_plural.endswith("s"):
                    verbose_name_plural += "s"

            # Get object verbose name
            object_name = cls._meta.verbose_name
            self.tag_model._meta.verbose_name = "%s %s" % (
                object_name,
                verbose_name_singular,
            )
            self.tag_model._meta.verbose_name_plural = "%s %s" % (
                object_name,
                verbose_name_plural,
            )

            # Make no attempt to enforce max length of verbose_name - no good
            # automatic solution, and the limit may change, see
            #   https://code.djangoproject.com/ticket/17763
            # If it's a problem, contrib.auth will raise a ValidationError

        #
        # Build the tag field
        #

        # Update the rel on the field
        self.remote_field.model = self.tag_model

        # Contribute to class
        super(BaseTagField, self).contribute_to_class(cls, name)
        self.contributed = True

    def formfield(self, form_class, **kwargs):
        """
        Common actions for TagField and SingleTagField to set up a formfield
        """
        required = not self.blank
        if hasattr(self, "required"):
            required = self.required

        # Update tag options, if necessary
        tag_options = self.tag_options
        if "tag_options" in kwargs:
            new_options = kwargs.pop("tag_options")
            if not isinstance(new_options, TagOptions):
                new_options = TagOptions(**new_options)
            tag_options += new_options

        # Start off with defaults
        options = {
            # Arguments the TagField base (CharField) would expect
            "label": capfirst(self.verbose_name),
            "help_text": self.help_text,
            "required": required,
            # Also pass tag options
            "tag_options": tag_options,
        }

        # Update with kwargs
        options.update(kwargs)

        # Add in list of tags for autocomplete, if appropriate
        if "autocomplete_tags" in kwargs:
            options["autocomplete_tags"] = kwargs["autocomplete_tags"]
        elif not tag_options.autocomplete_view:
            tags = self.tag_model.objects.all()
            if tag_options.autocomplete_initial:
                tags = tags.initial()
            options["autocomplete_tags"] = tags

        # Create the field instance
        return form_class(**options)

    def get_manager_name(self):
        """
        Get the field name for the Manager
        """
        return constants.TAGGED_ATTR_MANAGER % self.name

    def deconstruct(self):
        """
        Deconstruct field options to a dict for __init__
        """
        name, path, args, kwargs = super(BaseTagField, self).deconstruct()

        # Find tag model options
        if isinstance(self.tag_options, TagOptions):
            # When deconstruct is called on a real field on a real model,
            # we will have a concrete tag model, so use its tag options
            items = self.tag_options.items(with_defaults=False)

            # Freeze initial as a string, not array
            if "initial" in items:
                items["initial"] = self.tag_options.initial_string

        elif self._deferred_options is not None:
            # When deconstruct is called on a ModelState field, the options
            # will have been deferred
            set_tag_meta, options = self._deferred_options
            items = options

        else:  # pragma: no cover
            # It should never get here - raise an exception so we can debug
            raise ValueError("Unexpected state")

        # Add tag model options to kwargs
        kwargs.update(items)

        # Can't freeze lambdas. It should work if it's a function at the top
        # level of a module, but there's no easy way to differentiate. For
        # safety and consistency, strip callable arguments and mention in
        # documentation.
        if "get_absolute_url" in kwargs:
            del kwargs["get_absolute_url"]

        # Remove forbidden fields
        # This shouldn't be needed, but strip them just in case
        for forbidden in self.forbidden_fields:  # pragma: no cover
            if forbidden in kwargs:
                del kwargs[forbidden]

        # Always store _set_tag_meta=True, so migrating tag fields can set tag
        # models' TagMeta
        kwargs["_set_tag_meta"] = True

        return name, path, args, kwargs


# ##############################################################################
# ###### Single tag field
# ##############################################################################


class SingleTagField(BaseTagField, models.ForeignKey):
    """
    Build the tag model and register the TagForeignKey
    Not actually a field - syntactic sugar for creating tag fields
    """

    description = "A single tag field"

    # List of fields which are forbidden from __init__
    forbidden_fields = ("to_field", "rel_class", "max_count")

    def __init__(self, *args, **kwargs):
        """
        Create a single tag field - a tag field which can only take one tag

        See docs/models.rst for a list of arguments
        """
        # Forbid certain ForeignKey arguments
        for forbidden in self.forbidden_fields:
            if forbidden in kwargs:
                raise ValueError("Invalid argument '%s' for SingleTagField" % forbidden)

        # TagFields will need to be nulled in the database when deleting,
        # regardless of whether we want to allow them to be null or not.
        # Make a note of whether this is required.
        self.required = not kwargs.pop("blank", False)
        kwargs["blank"] = True
        kwargs["null"] = True

        # Set default on_delete
        if "on_delete" not in kwargs:
            kwargs["on_delete"] = models.CASCADE

        # Create the tag field
        super(SingleTagField, self).__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name):
        """
        Create the related tag model, initialise the ForeignKey with it,
        and set up the model to use the SingleTagDescriptor instead of the
        ReverseSingleRelatedObjectDescriptor
        """
        # Standard BaseTagField contribute
        super(SingleTagField, self).contribute_to_class(cls, name)

        # Replace the descriptor with our own
        old_descriptor = getattr(cls, name)
        new_descriptor = SingleTagDescriptor(old_descriptor)
        setattr(cls, name, new_descriptor)

    def value_from_object(self, obj):
        """
        Returns the value of the foreign key as a tag string, for a form
        """
        tag = getattr(obj, self.name)
        if tag:
            return tag.name
        return ""

    def formfield(self, form_class=forms.SingleTagField, **kwargs):
        """
        Create the form field
        For arguments see forms.TagField
        """
        return super(SingleTagField, self).formfield(form_class=form_class, **kwargs)


# ##############################################################################
# ###### Tag field
# ##############################################################################


class TagField(BaseTagField, models.ManyToManyField):
    """
    Build the tag model and register the TagManyToManyField
    Not actually a field - syntactic sugar for creating tag fields
    Will not allow a through table
    """

    description = "A tag field"

    # List of fields which are forbidden from __init__
    forbidden_fields = ("db_table", "through", "symmetrical")

    def __init__(self, *args, **kwargs):
        """
        Create a Tag field

        See docs/models.rst for a list of arguments
        """
        # Forbid certain ManyToManyField arguments
        for forbidden in self.forbidden_fields:
            if forbidden in kwargs:
                raise ValueError("Invalid argument '%s' for TagField" % forbidden)

        # Note things we'll need to restore after __init__
        help_text = kwargs.pop("help_text", "")
        null = kwargs.pop("null", False)

        super(TagField, self).__init__(*args, **kwargs)

        # Change default help text
        self.help_text = help_text or "Enter a comma-separated tag string"
        self.null = null

    def _check_ignored_options(self, **kwargs):
        warnings = []

        if getattr(self, "null", False):
            warnings.append(
                ChecksWarning(
                    "null has no effect on TagField.",
                    hint=None,
                    obj=self,
                    id="fields.W340",
                )
            )
            self.has_null_arg = False
            self.null = None

        return warnings + super(TagField, self)._check_ignored_options(**kwargs)

    def contribute_to_class(self, cls, name):
        """
        Create the related tag model, initialise the ManyToManyField with it,
        and set up the model to use the TagDescriptor
        """
        # Standard BaseTagField contribute
        super(TagField, self).contribute_to_class(cls, name)

        # Replace the descriptor with our own
        old_descriptor = getattr(cls, name)
        new_descriptor = TagDescriptor(old_descriptor)
        setattr(cls, name, new_descriptor)

    def value_from_object(self, obj):
        """
        Tricks django.forms.models.model_to_dict into passing data to the form.

        Because the models.TagField is based on a django ManyToManyField,
        model_to_dict expects a queryset, which it changes to a list of pks
        for use in a ModelMultipleChoiceField.

        Instead, we want to pass the tag string to a forms.TagField, which is a
        subclass of CharField. We can't do this in the TagField itself because
        there's no model context by that stage to do the pk lookup.

        We therefore return a fake queryset containing a single fake item,
        where the pk attribute is the tag string - a bit of a hack, but avoids
        monkey-patching Django.
        """

        class FakeObject(object):
            """
            FakeObject so m2d can check obj.pk
            """

            def __init__(self, value):
                self.pk = value

            def __str__(self):
                return self.pk

        class FakeQuerySet(object):
            """
            FakeQuerySet so m2d can call qs.values_list()
            Only contains one FakeObject instance
            """

            def __init__(self, obj):
                self.obj = obj
                self._result_cache = None

            def __iter__(self):
                """
                Iterable so m2d can use in list comprehension
                """
                yield self.obj

            def __len__(self):
                return 1

            def __getitem__(self, key):
                return self.obj

            def values_list(self, *fields, **kwargs):
                """
                Ignores arguments and returns an empty list with the object.pk
                """
                return [self.obj.pk]

        return FakeQuerySet(FakeObject(getattr(obj, self.attname).get_tag_string()))

    def formfield(self, form_class=forms.TagField, **kwargs):
        """
        Create the form field
        For arguments see forms.TagField
        """
        return super(TagField, self).formfield(form_class=form_class, **kwargs)

    def save_form_data(self, instance, data):
        """
        When the form wants to save this data, it will only be doing so after
        the instance has been saved. This must be committed to the database.
        """
        setattr(instance, self.attname, data)
        getattr(instance, self.attname).save()


# ##############################################################################
# ###### Field util methods
# ##############################################################################


def singletagfields_from_model(model):
    """
    Get a list of SingleTagField fields from a model class
    """
    return [field for field in model._meta.fields if isinstance(field, SingleTagField)]


def tagfields_from_model(model):
    """Get a list of TagField fields from a model class"""
    return [field for field in model._meta.many_to_many if isinstance(field, TagField)]



"""
Signals registered before models are loaded

These are registered as part of Tagulous model initialisation, so apps following this in
INSTALLED_APPS can have their models enhanced.
"""
from django.db.models.signals import class_prepared

from .. import settings
from ..models.tagged import TaggedModel


def class_prepared_listener(sender, **kwargs):
    """
    Listen to the class_prepared signal and subclass any model with tag
    fields
    """
    TaggedModel.cast_class(sender)


def register_pre_signals():
    """
    Called from tagulous/models/__init__.py
    """
    if settings.ENHANCE_MODELS:
        class_prepared.connect(class_prepared_listener, weak=False)
